import sqlalchemy

from enum import Enum
from jjsystem.database import db
from jjsystem.common.subsystem import entity
from sqlalchemy import orm


class FILE_SYNC_STATUS(Enum):
    PENDING = 1
    SUCCESS = 2
    ERROR = 3


class FILE_SYNC_MESSAGE_TYPE(Enum):
    INFO = 1
    WARNING = 2
    ERROR = 3


class FileSync(entity.Entity, db.Model):
    attributes = ['domain_id', 'file_id', 'status', 'entity', 'entity_id']
    attributes += entity.Entity.attributes

    domain_id = db.Column(
        db.CHAR(32), db.ForeignKey("domain.id"), nullable=False)
    file_id = db.Column(
        db.CHAR(32), db.ForeignKey("file_infosys.id"), nullable=False)
    status = db.Column(sqlalchemy.Enum(FILE_SYNC_STATUS), nullable=False)
    entity = db.Column(db.String(80), nullable=False)
    entity_id = db.Column(db.String(32), nullable=True)

    messages = orm.relationship('FileSyncMessage',
                                backref=orm.backref('file_sync_message'),
                                cascade='delete,delete-orphan,save-update')

    def __init__(self, id, domain_id, file_id, entity, status, entity_id=None,
                 active=True, created_at=None, created_by=None,
                 updated_at=None, updated_by=None, tag=None):
        super().__init__(id, active, created_at, created_by,
                         updated_at, updated_by, tag)
        self.domain_id = domain_id
        self.file_id = file_id
        self.status = status
        self.entity_id = entity_id

    @classmethod
    def embedded(cls):
        return ['messages']


class FileSyncMessage(entity.Entity, db.Model):
    attributes = ['domain_id', 'file_id']
    attributes += entity.Entity.attributes

    file_sync_id = db.Column(
        db.CHAR(32), db.ForeignKey('file_sync.id'), nullable=False)
    type_message = db.Column(sqlalchemy.Enum(FILE_SYNC_MESSAGE_TYPE), nullable=False)
    body = db.Column(db.String(250), nullable=False)

    def __init__(self, id, file_sync_id, type_message, body,
                 active=True, created_at=None, created_by=None,
                 updated_at=None, updated_by=None, tag=None):
        super().__init__(id, active, created_at, created_by,
                         updated_at, updated_by, tag)
        self.file_sync_id = file_sync_id
        self.type_message = type_message
        self.body = body



from copy import copy
from typing import Tuple, List

InputResource = Tuple[str, List[str]]
RouteResource = Tuple[str, str]


class InputResourceUtils:

    @classmethod
    def parse_resources(cls, resources: List[InputResource]) \
            -> List[RouteResource]:
        output_resources = []
        for (endpoint, methods) in resources:
            for method in methods:
                output_resources.append((endpoint, method))
        return cls.remove_duplicates(output_resources)

    @classmethod
    def remove_duplicates(cls, resources: List[RouteResource]) \
            -> List[RouteResource]:
        return list(set([resource for resource in resources]))

    @classmethod
    def diff_resources(cls, first: List[RouteResource],
                       *args: List[RouteResource]):
        temp = set(copy(first))
        for arg in args:
            temp -= set(arg)
        return list(temp)

    @classmethod
    def join_resources(cls, first: List[RouteResource],
                       *args: List[RouteResource]):
        temp = copy(first)
        for arg in args:
            temp += arg
        return cls.remove_duplicates(temp)



from jjsystem.common.subsystem import router


class Router(router.Router):

    def __init__(self, collection, routes=[]):
        super().__init__(collection, routes)

    @property
    def routes(self):
        return super().routes + [
            {
                'action': 'Adicionar routes',
                'method': 'PUT',
                'url': self.resource_url + '/add_routes',
                'callback': 'add_routes'
            },
            {
                'action': 'Remove routes',
                'method': 'PUT',
                'url': self.resource_url + '/rm_routes',
                'callback': 'rm_routes'
            },
            {
                'action': 'Pega as rotas disponiveis',
                'method': 'GET',
                'url': self.resource_url + '/get_available_routes',
                'callback': 'get_available_routes'
            },
            {
                'action': 'Pega as rotas selecionadas',
                'method': 'GET',
                'url': self.resource_url + '/get_selected_routes',
                'callback': 'get_selected_routes'
            }
        ]



from jjsystem.common import exception
from jjsystem.common.subsystem import operation, manager


class GetNextVal(operation.Operation):
    def pre(self, session, id, name, **kwargs):
        domain = self.manager.api.domains().get(id=id)
        if not domain:
            raise exception.NotFound('ERROR! Domain not found')
        self.domain_id = domain.id

        if not name:
            raise exception.BadRequest('ERROR! Invalid name')
        self.name = name

        return True

    def do(self, session, **kwargs):
        nextval = self.driver.get_nextval(session, self.domain_id, self.name)

        if nextval is None:
            raise exception.JJSystemException(
                'Was not possible retrive next val')

        return nextval


class Manager(manager.Manager):

    def __init__(self, driver):
        super(Manager, self).__init__(driver)
        self.get_nextval = GetNextVal(self)



import flask
import uuid
import os
from werkzeug import utils as werkzeug_utils

from jjsystem.common.exception import BadRequest
from jjsystem.common.subsystem import manager
from jjsystem.common.subsystem import operation


# TODO(samueldmq): put this in the app config
UPLOAD_FOLDER = 'uploads'


class Create(operation.Create):

    def __call__(self, file, **kwargs):
        self.file = file
        self.domain_id = kwargs.get('domain_id', None)
        self.user_id = kwargs.pop('user_id', None)
        self.is_internal = type(self.file).__name__ == 'BufferedWriter' \
            or type(self.file).__name__ == 'BufferedReader'
        self.is_reader = type(self.file).__name__ == 'BufferedReader'
        self.filename = self.file.name \
            if self.is_internal else self.file.filename
        return super().__call__(**kwargs)

    def pre(self, session, **kwargs):
        if self.file and self.manager.allowed_file(self.filename):
            filename = werkzeug_utils.secure_filename(self.filename)

            if 'id' not in kwargs.keys():
                kwargs['id'] = uuid.uuid4().hex
            kwargs['created_by'] = self.user_id
            kwargs['name'] = filename
            # TODO note that kwargs must contain olny entity fields
            self.entity = self.driver.instantiate(**kwargs)
        else:
            e = BadRequest()
            e.message = 'file not allowed'
            raise e
        self.upload_folder = self.manager.get_upload_folder(self.entity,
                                                            self.domain_id)
        return self.entity.is_stable()

    def do(self, session, **kwargs):
        if not self.is_internal:
            self.file.save(os.path.join(
                self.upload_folder, self.entity.filename))
        elif self.is_reader:
            with open(
                os.path.join(
                    self.upload_folder, self.entity.filename), 'wb') as file:
                file.write(self.file.read())
        entity = super().do(session)
        return entity


class Get(operation.Get):

    def do(self, session, **kwargs):
        file = super().do(session, **kwargs)

        folder = self.manager.get_upload_folder(file, file.domain_id)
        # folder = get_upload_folder(file.domain_id)
        return flask.send_from_directory(folder, file.filename)


class Manager(manager.Manager):

    OPTIONAL_FOLDER = ''
    ALLOWED_EXTENSIONS = ['txt', 'pdf', 'png', 'jpg', 'jpeg', 'gif', 'csv']

    def __init__(self, driver):
        super().__init__(driver)
        self.create = Create(self)
        self.get = Get(self)

    @classmethod
    def allowed_file(cls, filename):
        return '.' in filename and \
               filename.rsplit('.', 1)[1] in cls.ALLOWED_EXTENSIONS

    @classmethod
    def get_upload_folder(cls, entity, domain_id):
        base_folder = cls._get_base_folder()
        entity_name = type(entity).__name__
        folder = os.path.join(base_folder,
                              entity_name,
                              cls.OPTIONAL_FOLDER,
                              domain_id)
        if not os.path.exists(folder):
            os.makedirs(folder)
        return folder

    @classmethod
    def _get_base_folder(cls):
        env_folder = os.environ.get('JJSYSTEM_FILE_DIR', UPLOAD_FOLDER)
        if not os.path.isabs(env_folder):
            env_folder = os.path.join(os.getcwd(), env_folder)
        return env_folder



from jjsystem.common.subsystem.apihandler import Api
from typing import Dict

from jjsystem.common import utils
from jjsystem.subsystem.application.resource import Application


class BootstrapApplication(object):

    def __init__(self, api: Api) -> None:
        self.application_manager = api.applications()

    def execute(self) -> Application:
        application = self._get_application_default()
        return self._save_application(application)

    def _get_application_default(self) -> Application:
        return Application(id=utils.random_uuid(),
                           name=Application.DEFAULT,
                           description="Application Default")

    def _save_application(self, application: Application) -> Application:
        data = application.to_dict()
        if 'settings' in data.keys():
            data.pop('settings')
        return self.application_manager.create(**data)



from jjsystem.common.subsystem import router


class Router(router.Router):

    def __init__(self, collection, routes=[]):
        super().__init__(collection, routes)

    @property
    def routes(self):
        settings_endpoint = '/settings'
        return super().routes + [
            {
                'action': 'restore',
                'method': 'POST',
                'url': self.collection_url + '/restore',
                'callback': 'restore',
                'bypass': True
            },
            {
                'action': 'reset_password',
                'method': 'POST',
                'url': self.resource_url + '/reset',
                'callback': 'reset_password'
            },
            {
                'action': 'reset_my_password',
                'method': 'POST',
                'url': self.collection_url + '/reset',
                'callback': 'reset_my_password'
            },
            {
                'action': 'update_my_password',
                'method': 'PUT',
                'url': self.resource_url + '/update_my_password',
                'callback': 'update_password'
            },
            {
                'action': 'routes',
                'method': 'GET',
                'url': self.collection_url + '/routes',
                'callback': 'routes'
            },
            {
                'action': 'routes',
                'method': 'PUT',
                'url': self.resource_url + '/photo',
                'callback': 'upload_photo'
            },
            {
                'action': 'routes',
                'method': 'DELETE',
                'url': self.resource_url + '/photo',
                'callback': 'delete_photo'
            },
            {'action': 'notify', 'method': 'POST',
                'url': self.resource_url + '/notify',
                'callback': 'notify'},
            {
                'action': 'roles',
                'method': 'GET',
                'url': self.resource_url + '/roles',
                'callback': 'roles'
            },
            {
                'action': 'Update settings on User',
                'method': 'PUT',
                'url': self.resource_url + settings_endpoint,
                'callback': 'update_settings',
                'bypass': False
            },
            {
                'action': 'Remove settings from User',
                'method': 'DELETE',
                'url': self.resource_url + settings_endpoint,
                'callback': 'remove_settings',
                'bypass': False
            },
            {
                'action': 'Get settings by keys from User',
                'method': 'GET',
                'url': self.resource_url + settings_endpoint,
                'callback': 'get_user_settings_by_keys',
                'bypass': False
            }
        ]



import flask

from jjsystem.common.subsystem import controller
from jjsystem.common import exception


class Controller(controller.Controller):

    def __init__(self, manager, resource_wrap, collection_wrap):
        super(Controller, self).__init__(
            manager, resource_wrap, collection_wrap)

    def deletar_tokens(self):
        data = self._filters_parse()
        try:
            self.manager.deletar_tokens(**data)
        except exception.JJSystemException as exc:
            return flask.Response(response=exc.message,
                                  status=exc.status)

        return flask.Response(response=None,
                              status=204,
                              mimetype="application/json")



from typing import List

from jjsystem.common.subsystem import Subsystem
from jjsystem.common.input import InputResource, InputResourceUtils


class System(object):

    def __init__(self, name: str,
                 subsystems: List[Subsystem],
                 user_resources: List[InputResource] = [],
                 sysadmin_resources: List[InputResource] = [],
                 sysadmin_exclusive_resources: List[InputResource] = []):
        self.name = name
        self.subsystems = subsystems

        parse_resources = InputResourceUtils.parse_resources
        self.user_resources = parse_resources(user_resources)
        self.sysadmin_exclusive_resources = \
            parse_resources(sysadmin_exclusive_resources)
        self.sysadmin_resources = parse_resources(sysadmin_resources)
        self.sysadmin_resources = InputResourceUtils.\
            join_resources(self.sysadmin_resources,
                           self.sysadmin_exclusive_resources)



from sqlalchemy import and_
from jjsystem.common import exception
from jjsystem.common.subsystem.pagination import Pagination
from jjsystem.subsystem.application.resource import Application
from jjsystem.subsystem.capability.resource import Capability
from jjsystem.subsystem.capability_functionality.resource \
    import CapabilityFunctionality
from jjsystem.subsystem.capability_module.resource import CapabilityModule
from jjsystem.subsystem.domain import tasks
from jjsystem.common.subsystem import operation, manager
from jjsystem.subsystem.domain.resource import Domain
from jjsystem.subsystem.functionality.resource import FunctionalityRoute
from jjsystem.subsystem.image.resource import QualityImage
from jjsystem.subsystem.domain.files import utils
from jjsystem.common.utils import round_abnt
from jjsystem.subsystem.module.resource import ModuleFunctionality
from jjsystem.subsystem.policy.resource import Policy
from jjsystem.subsystem.role.resource import Role
from sqlalchemy.orm import aliased


class DomainByName(operation.Operation):

    def pre(self, session, domain_name, **kwargs):
        self.domain_name = domain_name
        return True

    def do(self, session, **kwargs):
        domains = self.manager.list(name=self.domain_name)
        if not domains:
            raise exception.NotFound('ERROR! Domain name not found')
        domain = domains[0]

        # Hide user ID and settings from public resources
        domain.created_by = None
        domain.updated_by = None

        return domain


class DomainLogoByName(operation.Operation):

    def pre(self, session, domain_name, **kwargs):
        self.domain_name = domain_name
        return True

    def do(self, session, **kwargs):
        domains = self.manager.list(name=self.domain_name)
        if not domains:
            raise exception.NotFound('ERROR! Domain name not found')
        domain = domains[0]

        if domain.logo_id is None:
            raise exception.NotFound('ERROR! Domain logo not found')

        kwargs['quality'] = kwargs.get('quality', QualityImage.med)
        return self.manager.api.images().get(id=domain.logo_id, **kwargs)


class UploadLogo(operation.Update):

    def pre(self, session, id, token, file, **kwargs):
        self.file = file
        self.token = token

        return super().pre(session, id, **kwargs)

    def do(self, session, **kwargs):
        kwargs = {}
        kwargs['domain_id'] = self.entity.id
        kwargs['user_id'] = self.token.user_id
        kwargs['type_image'] = 'DomainLogo'

        image = self.manager.api.images().create(file=self.file, **kwargs)

        self.entity.logo_id = image.id

        return super().do(session=session)


class RemoveLogo(operation.Update):

    def do(self, session, **kwargs):
        logo_id = self.entity.logo_id
        self.entity.logo_id = None
        entity = super().do(session=session)
        if logo_id and entity is not None:
            self.manager.api.images().delete(id=logo_id)

        return


class Register(operation.Create):

    def _register_domain(self, domain_name, domain_display_name,
                         application_id, username, email, password):
        self.domain = self.manager.api.domains().create(
            application_id=application_id, name=domain_name,
            display_name=domain_display_name,
            addresses=[], contacts=[], active=False)

        self.user = self.manager.api.users().create(
            name=username, email=email, domain_id=self.domain.id, active=False)

        self.manager.api.users().reset(id=self.user.id, password=password)

    def pre(self, session, username, email, password,
            domain_name, domain_display_name, application_name):
        self.username = username
        self.email = email
        self.password = password
        self.domain_name = domain_name
        self.domain_display_name = domain_display_name
        self.application_name = application_name

        user_params_ok = username and email and password
        app_params_ok = domain_name and application_name
        if not user_params_ok and app_params_ok:
            raise exception.BadRequest(
                'ERROR! Not enough data to register domain')

        applications = \
            self.manager.api.applications().list(name=application_name)
        if not applications:
            raise exception.BadRequest('ERROR! Application name not found.')
        self.application = applications[0]

        return True

    def do(self, session, **kwargs):
        domains = self.manager.api.domains().list(name=self.domain_name)
        if not domains:
            self._register_domain(
                self.domain_name, self.domain_display_name,
                self.application.id, self.username, self.email, self.password)
        else:
            self.domain = domains[0]
            domain = domains[0]

            users = self.manager.api.users().list(email=self.email,
                                                  domain_id=domain.id)

            if domain.active:
                raise exception.BadRequest('Domain already activated')

            if not users or domain.display_name != self.domain_display_name:
                raise exception.BadRequest('Domain already registered')

            self.user = users[0]
            self.manager.api.users().reset(id=self.user.id,
                                           password=self.password)

        return True

    def post(self):
        # The notification don't be part of transaction must be on post
        tasks.send_email(self.user.id)


class Activate(operation.Create):

    def pre(self, session, token_id, domain_id, user_admin_id):
        if not (user_admin_id or domain_id):
            raise exception.BadRequest(
                'ERROR! Not enough data to Activate Domain')

        self.token_id = token_id
        self.domain_id = domain_id
        self.user_admin_id = user_admin_id

        roles = self.manager.api.roles().list(name='Admin')
        if not roles:
            raise exception.BadRequest('ERROR! Role Admin not found')
        self.role_admin = roles[0]

        return True

    def do(self, session, **kwargs):
        self.manager.api.domains().update(id=self.domain_id, active=True)
        self.manager.api.users().update(id=self.user_admin_id, active=True)
        self.manager.api.grants().create(user_id=self.user_admin_id,
                                         role_id=self.role_admin.id)
        self.manager.api.tokens().delete(id=self.token_id)

        domain = self.manager.api.domains().get(id=self.domain_id)
        if not domain:
            raise exception.BadRequest('ERROR! Domain not found')

        return domain


class UpdateSettings(operation.Update):

    def pre(self, session, id: str, **kwargs) -> bool:
        self.settings = kwargs
        if self.settings is None or not self.settings:
            raise exception.BadRequest("Erro! There is not a setting")
        return super().pre(session=session, id=id)

    def do(self, session, **kwargs):
        result = {}
        for key, value in self.settings.items():
            new_value = self.entity.update_setting(key, value)
            result[key] = new_value
        super().do(session)

        return result


class RemoveSettings(operation.Update):

    def pre(self, session, id: str, **kwargs) -> bool:
        self.keys = kwargs.get('keys', [])
        if not self.keys:
            raise exception.BadRequest('Erro! keys are empty')
        super().pre(session, id=id)

        return self.entity.is_stable()

    def do(self, session, **kwargs):
        result = {}
        for key in self.keys:
            value = self.entity.remove_setting(key)
            result[key] = value
        super().do(session=session)

        return result


class GetDomainSettingsByKeys(operation.Get):

    def pre(self, session, id, **kwargs):
        self.keys = kwargs.get('keys', [])
        if not self.keys:
            raise exception.BadRequest('Erro! keys are empty')
        return super().pre(session, id=id)

    def do(self, session, **kwargs):
        entity = super().do(session=session)
        settings = {}
        for key in self.keys:
            value = entity.settings.get(key, None)
            if value is not None:
                settings[key] = value
        return settings


class SendEmailActivateAccount(operation.Create):

    def pre(self, session, **kwargs):
        return True

    def do(self, session, **kwargs):
        domain_name = kwargs.get('domain_name', None)
        email = kwargs.get('email', None)
        domains = self.manager.api.domains().list(name=domain_name)
        if not domains:
            raise exception.NotFound('Domain not found')

        domain = domains[0]

        users = self.manager.api.users().list(email=email,
                                              domain_id=domain.id,
                                              name='admin')

        if domain.active:
            raise exception.BadRequest('Domain already activated')

        if not users:
            raise exception.NotFound('Invalid information')

        self.user = users[0]

        return True

    def post(self):
        # The notification don't be part of transaction must be on post
        tasks.send_email(self.user.id)


class GetFilesSize(operation.Get):

    def pre(self, session, id, **kwargs):
        return super().pre(session, id=id)

    def do(self, session, **kwargs):
        response_dict = utils.get_size(domain_id=self.id, **kwargs)
        return response_dict


class GetUsageInfoByDomain(operation.List):

    PRECISION = 2

    # vai pegar os totais de linhas em cada tabela do banco por domínio
    def _get_totals_by_domain(self, session):
        msg = """
            É necessário implementar a função get_totals_by_domain no
            projeto específico para garantir que as tabelas listadas e
            filtradas por domínio são apenas as importantes. O retorno
            deve ser uma lista de dicionários no seguinte formato:

            return = [
                {
                    'table_name': None,
                    'total': None,
                    'domain_name': None,
                    'application_id': None,
                    'application_name': None
                    'domain_id': None,
                    'active': None
                }
            ]
        """
        raise exception.BadRequest(msg)

    # vai pegar os totais de linhas em cada tabela do banco
    def _get_totals_db(self, session):
        msg = """
            É necessário implementar a função _get_totals_db no
            projeto específico para garantir que as tabelas listadas
            são apenas as importantes para aquele projeto. O retorno
            deve ser uma lista de dicionários no seguinte formato:

            return = {
                "entity_name1": total1,
                "entity_name2": total2,
                "entity_name3": total3
            }
        """
        raise exception.BadRequest(msg)

    def _normalizar_tamanho(self, tamanho: str):
        numero, unidade = tamanho.split(' ')
        numero = int(numero)
        if unidade == 'bytes':
            return numero
        elif unidade == 'kB':
            return (numero * 1024)
        elif unidade == 'MB':
            return (numero * 1024 * 1024)
        elif unidade == 'GB':
            return (numero * 1024 * 1024 * 1024)

    def _get_value_or_default(self, value, tipo):
        if value is not None:
            return value
        else:
            if tipo is str:
                return ''
            elif tipo is int:
                return 0

    # vai pegar o total de espaço de disco ocupado por cada tabela do banco
    def _get_disk_space_totals_by_db(self, session):
        from jjsystem.subsystem.domain.sql.disk_space_totals_by_db \
            import QUERY
        rs = session.execute(QUERY)
        response = [r for r in rs]

        response_dict = {
            r[0]: self._normalizar_tamanho(r[1])
            for r in response}
        return response_dict

    # função responsável por pegar o objeto constante com as constantes para
    # realizar o cálculo
    def _get_const(self, application_id, table_name):
        for const in self.constants:
            if const.application_id == application_id and\
               const.table_name == table_name:
                return (const.p_processos, const.p_acessos_db)
        return (1, 1)

    # vai pegar as "constantes" que serão usadas para aplicar os pesos
    # de cada cálculo
    def _get_constant_for_calculation(self):
        self.constants, _ = self.manager.api.constant_for_calculations().list(
            active=True)

    def _apply_calculations(self):
        for i in range(0, len(self.totals_by_domain)):
            total = self.totals_by_domain[i].get('total', 0)
            table_name = self.totals_by_domain[i].get('table_name', '')
            # domain_name = self.totals_by_domain[i].get('domain_name', '')
            application_id = self.totals_by_domain[i].get(
                'application_id', '')
            percent = (total / self.totals_db.get(table_name, 1))
            disk_space_table = self.disk_space_totals_by_db_dict.get(
                table_name, 0)

            self.totals_by_domain[i]['percent'] = percent
            self.totals_by_domain[i]['total_space_disc_bank'] = (
                percent * disk_space_table)

            p_processos, p_acessos_db = self._get_const(
                application_id, table_name)
            self.totals_by_domain[i]['total_process'] = total * p_processos
            self.totals_by_domain[i]['total_acess_db'] = total * p_acessos_db

    def _make_base_dict(self, domain_name):
        return {
            'application_id': None,
            'application_name': None,
            'domain_id': None,
            'domain_name': domain_name,
            'total_entities': 0,
            'total_space_disc_bank': 0,
            'total_process': 0,
            'total_acess_db': 0,
            'total_space_disc_file': 0
        }

    def _group_by_domain(self):
        grand_totals_dict = {
            'total_entities': 0,
            'total_space_disc_bank': 0,
            'total_process': 0,
            'total_acess_db': 0,
            'total_space_disc_file': 0
        }

        domains = list(
            set(map(lambda x: x.get('domain_name', ''), self.totals_by_domain)))
        domains_dict = list(map(lambda x: self._make_base_dict(x), domains))
        response = []
        for dc in domains_dict:
            domain_name = dc.get('domain_name', '')
            list_aux = list(filter(lambda x: x['domain_name'] == domain_name,
                                   self.totals_by_domain))

            # preenche os campos do dicionário que estavam faltando
            dc['application_name'] = list_aux[0].get('application_name', None)
            dc['application_id'] = list_aux[0].get('application_id', None)
            dc['domain_id'] = list_aux[0].get('domain_id', None)
            dc['active'] = list_aux[0].get('active', False)

            # soma todos os totais de um domínio
            for aux in list_aux:
                total_entities = aux.get('total', 0)
                total_space_disc_bank = aux.get('total_space_disc_bank', 0)
                total_process = aux.get('total_process', 0)
                total_acess_db = aux.get('total_acess_db', 0)

                dc['total_entities'] += total_entities
                dc['total_space_disc_bank'] += total_space_disc_bank
                dc['total_process'] += total_process
                dc['total_acess_db'] += total_acess_db

            # calcula o espaço ocupado por arquivos no servidor para este
            # domínio
            dc['total_space_disc_file'] = self.manager.get_files_size(
                id=dc['domain_id'], de=self.de, ate=self.ate).get('BYTES', 0)

            # faz a somatória total de todos os campos
            grand_totals_dict['total_entities'] += dc['total_entities']
            grand_totals_dict['total_space_disc_bank'] += \
                dc['total_space_disc_bank']
            grand_totals_dict['total_process'] += dc['total_process']
            grand_totals_dict['total_acess_db'] += dc['total_acess_db']
            grand_totals_dict['total_space_disc_file'] += \
                dc['total_space_disc_file']

            response.append(dc)

        return (response, grand_totals_dict)

    def _calculate_cost_individual(self, usage, cost):
        return usage * float(cost)

    def _calculate_cost(self, usage):
        return ((usage * float(self.cost_servidor)) +
                (usage * float(self.cost_operacional)) +
                (usage * float(self.cost_administrativo)))

    def _calculate_usage(
            self, percent_total_entities, percent_total_space_disc_bank,
            percent_total_process, percent_total_access_db,
            percent_total_space_disc_file):
        usage_storage_bank = (
            float(percent_total_space_disc_bank) * self.p_storage_bank)
        usage_storage_file = (
            float(percent_total_space_disc_file) * self.p_storage_file)
        usage_process = float(percent_total_process) * self.p_process
        usage_access_db = float(percent_total_access_db) * self.p_access_db

        return (usage_storage_bank + usage_storage_file + usage_process +
                usage_access_db)

    def _normalize_values_response(self, response_item):
        response_item['total_space_disc_bank'] = round_abnt(
            response_item['total_space_disc_bank'], self.PRECISION)
        response_item['total_process'] = round_abnt(
            response_item['total_process'], self.PRECISION)
        response_item['total_acess_db'] = round_abnt(
            response_item['total_acess_db'], self.PRECISION)
        response_item['total_space_disc_file'] = round_abnt(
            response_item['total_space_disc_file'], self.PRECISION)
        response_item['percent_total_entities'] = round_abnt(
            response_item['percent_total_entities'], self.PRECISION)
        response_item['percent_total_space_disc_bank'] = round_abnt(
            response_item['percent_total_space_disc_bank'], self.PRECISION)
        response_item['percent_total_process'] = round_abnt(
            response_item['percent_total_process'], self.PRECISION)
        response_item['percent_total_access_db'] = round_abnt(
            response_item['percent_total_access_db'], self.PRECISION)
        response_item['percent_total_space_disc_file'] = round_abnt(
            response_item['percent_total_space_disc_file'], self.PRECISION)
        response_item['usage'] = round_abnt(
            response_item['usage'], self.PRECISION)
        response_item['cost_servidor'] = round_abnt(
            response_item['cost_servidor'], self.PRECISION)
        response_item['cost_operacional'] = round_abnt(
            response_item['cost_operacional'], self.PRECISION)
        response_item['cost_administrativo'] = round_abnt(
            response_item['cost_administrativo'], self.PRECISION)
        response_item['cost'] = (
            response_item['cost_servidor'] +
            response_item['cost_operacional'] +
            response_item['cost_administrativo']
        )
        response_item['cost'] = round_abnt(
            response_item['cost'], self.PRECISION)

        return response_item

    def _normalize_values_grand_totals_dict(self, grand_totals_dict):
        grand_totals_dict['total_entities'] = round_abnt(
            grand_totals_dict['total_entities'], self.PRECISION)
        grand_totals_dict['total_space_disc_bank'] = round_abnt(
            grand_totals_dict['total_space_disc_bank'], self.PRECISION)
        grand_totals_dict['total_process'] = round_abnt(
            grand_totals_dict['total_process'], self.PRECISION)
        grand_totals_dict['total_acess_db'] = round_abnt(
            grand_totals_dict['total_acess_db'], self.PRECISION)
        grand_totals_dict['total_space_disc_file'] = round_abnt(
            grand_totals_dict['total_space_disc_file'], self.PRECISION)
        grand_totals_dict['cost_servidor'] = round_abnt(
            grand_totals_dict['cost_servidor'], self.PRECISION)
        grand_totals_dict['cost_operacional'] = round_abnt(
            grand_totals_dict['cost_operacional'], self.PRECISION)
        grand_totals_dict['cost_administrativo'] = round_abnt(
            grand_totals_dict['cost_administrativo'], self.PRECISION)

        return grand_totals_dict

    def fill_in_percentages(self, response, grand_totals_dict):
        total_entities = grand_totals_dict['total_entities']
        total_space_disc_bank = grand_totals_dict['total_space_disc_bank']
        total_process = grand_totals_dict['total_process']
        total_acess_db = grand_totals_dict['total_acess_db']
        total_space_disc_file = grand_totals_dict['total_space_disc_file']
        if total_space_disc_file == 0:
            total_space_disc_file = 1

        for i in range(0, len(response)):
            response[i]['percent_total_entities'] = (
                response[i]['total_entities'] / total_entities)
            response[i]['percent_total_space_disc_bank'] = (
                response[i]['total_space_disc_bank'] / total_space_disc_bank)
            response[i]['percent_total_process'] = (
                response[i]['total_process'] / total_process)
            response[i]['percent_total_access_db'] = (
                response[i]['total_acess_db'] / total_acess_db)
            response[i]['percent_total_space_disc_file'] = (
                response[i]['total_space_disc_file'] / total_space_disc_file)

            response[i]['usage'] = self._calculate_usage(
                response[i]['percent_total_entities'],
                response[i]['percent_total_space_disc_bank'],
                response[i]['percent_total_process'],
                response[i]['percent_total_access_db'],
                response[i]['percent_total_space_disc_file']
            )
            usage = response[i]['usage']
            response[i]['cost_servidor'] = self._calculate_cost_individual(
                usage, self.cost_servidor)
            response[i]['cost_operacional'] = self._calculate_cost_individual(
                usage, self.cost_operacional)
            response[i]['cost_administrativo'] = self\
                ._calculate_cost_individual(usage, self.cost_administrativo)
            response[i]['cost'] = self._calculate_cost(response[i]['usage'])
            response[i] = self._normalize_values_response(response[i])

        return response

    def _valid_weights(self):
        total = (self.p_storage_bank + self.p_storage_file +  # noqa
                 self.p_process + self.p_access_db)
        if total > 1:
            raise exception.BadRequest(
                'the sum of the general weights must total 100%')

    def _get_cost_db(self, session, cost_name):
        return self.manager.api.project_costs().\
            get_cost_by_name_most_recent(
                session=session, name=cost_name, active=True)

    def _get_costs_params(self, **kwargs):
        session = kwargs.get('session', None)
        self.cost_servidor = kwargs.pop(
            'cost_servidor',
            self._get_cost_db(session, 'servidor'))
        self.cost_operacional = kwargs.pop(
            'cost_operacional',
            self._get_cost_db(session, 'operacional'))
        self.cost_administrativo = kwargs.pop(
            'cost_administrativo',
            self._get_cost_db(session, 'administrativo'))

        if self.cost_servidor is None:
            self.cost_servidor = 0
        if self.cost_operacional is None:
            self.cost_operacional = 0
        if self.cost_administrativo is None:
            self.cost_administrativo = 0

    def pre(self, **kwargs):
        self.de = kwargs.get('de', '2020-01-01-0300')
        self.ate = kwargs.get('ate', '9020-01-01-0300')
        self.domain_name = kwargs.get('domain_name', None)

        session = kwargs.get('session', None)

        self.p_storage_bank = float(kwargs.get('p_storage_bank', 0.1))
        self.p_storage_file = float(kwargs.get('p_storage_file', 0.1))
        self.p_process = float(kwargs.get('p_process', 0.3))
        self.p_access_db = float(kwargs.get('p_access_db', 0.5))

        self._valid_weights()

        precision = kwargs.get('precision', None)
        if precision is not None:
            try:
                self.PRECISION = int(precision)
            except Exception:
                raise exception.BadRequest(
                    'the precision parameter must be an integer')

        self.totals_db = self._get_totals_db(session)
        self.disk_space_totals_by_db_dict = self._get_disk_space_totals_by_db(
            session)
        self._get_constant_for_calculation()
        self.totals_by_domain = self._get_totals_by_domain(session)

        self._get_costs_params(**kwargs)
        return True

    def do(self, session, **kwargs):
        self._apply_calculations()

        response, grand_totals_dict = self._group_by_domain()
        response = self.fill_in_percentages(response, grand_totals_dict)

        grand_totals_dict['cost_servidor'] = self.cost_servidor
        grand_totals_dict['cost_operacional'] = self.cost_operacional
        grand_totals_dict['cost_administrativo'] = self.cost_administrativo
        grand_totals_dict['p_storage_bank'] = self.p_storage_bank
        grand_totals_dict['p_storage_file'] = self.p_storage_file
        grand_totals_dict['p_process'] = self.p_process
        grand_totals_dict['p_access_db'] = self.p_access_db
        grand_totals_dict = self._normalize_values_grand_totals_dict(
            grand_totals_dict)
        return (response, grand_totals_dict)


class GetRoles(operation.Operation):

    def pre(self, session, id, **kwargs):
        self.entity = self.driver.get(id, session=session)
        return self.entity is not None

    def do(self, session, **kwargs):
        roles = session.query(Role). \
            join(Policy). \
            join(Capability). \
            join(Application). \
            filter(and_(Capability.application_id == self.entity.application_id,
                        Role.name != Role.USER)). \
            distinct()
        return roles


class CheckPermission(operation.Get):

    def pre(self, session, id, url, method, **kwargs):
        self.id = id
        routes = self.manager.api.routes().list(
            url=url, method=method)

        if not routes:
            msg = 'Route not found'
            raise exception.BadRequest(msg)
        self.route = routes[0]

        if not self.route.active:
            msg = 'Route is inactive'
            raise exception.BadRequest(msg)

        return True

    def do(self, session, **kwargs):
        domain = super().do(session, **kwargs)
        if self.route.bypass:
            # return (has_permission, route)
            return (True, self.route)
        elif self.route.sysadmin and domain.name != 'default':
            return (False, self.route)

        has_capabilities = session.query(Domain.id). \
            join(Application). \
            join(Capability,
                 and_(Capability.application_id == Application.id,
                      Capability.route_id == self.route.id)). \
            filter(and_(Domain.id == self.id,
                        Domain.active, Application.active,
                        Capability.active)). \
            count()

        has_capability_functionalities = session.query(Domain.id). \
            join(Application). \
            join(CapabilityFunctionality). \
            join(FunctionalityRoute,
                 and_(FunctionalityRoute.functionality_id
                      == CapabilityFunctionality.functionality_id,
                      FunctionalityRoute.route_id == self.route.id)). \
            filter(and_(Domain.id == self.id,
                        Domain.active, Application.active,
                        CapabilityFunctionality.active,
                        FunctionalityRoute.active)). \
            count()

        has_capability_modules = session.query(Domain.id). \
            join(Application). \
            join(CapabilityModule). \
            join(ModuleFunctionality,
                 ModuleFunctionality.module_id
                 == CapabilityModule.module_id). \
            join(FunctionalityRoute,
                 and_(FunctionalityRoute.functionality_id
                      == ModuleFunctionality.functionality_id,
                      FunctionalityRoute.route_id == self.route.id)). \
            filter(and_(Domain.id == self.id,
                        Domain.active, Application.active,
                        ModuleFunctionality.active,
                        FunctionalityRoute.active)). \
            count()

        total = (has_capabilities + has_capability_functionalities +
                 has_capability_modules)
        hsa_permission = total > 0

        return (hsa_permission, self.route)


class List(operation.List):

    def do(self, session, **kwargs):
        Domain2 = aliased(Domain)

        query = session.query(Domain). \
            join(Application, Application.id == Domain.application_id). \
            join(Domain2, Domain2.id == Domain.parent_id, isouter=True)

        dict_compare = {"application.": Application,
                        "parent.": Domain2}
        query = query.distinct()
        kwargs['query'] = query
        kwargs['dict_compare'] = dict_compare
        return super().do(session=session, **kwargs)


class Count(operation.Count):

    def do(self, session, **kwargs):
        Domain2 = aliased(Domain)

        query = session.query(Domain). \
            join(Application, Application.id == Domain.application_id). \
            join(Domain2, Domain2.id == Domain.parent_id, isouter=True)
        query = self.manager.apply_filters(query, Domain, **kwargs)

        dict_compare = {"application.": Application,
                        "parent.": Domain2}
        query = self.manager.apply_filters_includes(
            query, dict_compare, **kwargs)
        query = query.distinct()

        result = query.count()

        return result


class Manager(manager.Manager):

    def __init__(self, driver):
        super(Manager, self).__init__(driver)
        self.domain_by_name = DomainByName(self)
        self.domain_logo_by_name = DomainLogoByName(self)
        self.upload_logo = UploadLogo(self)
        self.remove_logo = RemoveLogo(self)
        self.register = Register(self)
        self.activate = Activate(self)
        self.update_settings = UpdateSettings(self)
        self.remove_settings = RemoveSettings(self)
        self.get_domain_settings_by_keys = GetDomainSettingsByKeys(self)
        self.send_email_activate_account = SendEmailActivateAccount(self)
        self.get_files_size = GetFilesSize(self)
        self.get_usage_info_by_domain = GetUsageInfoByDomain(self)
        self.get_roles = GetRoles(self)
        self.check_permission = CheckPermission(self)
        self.list = List(self)
        self.count = Count(self)



from typing import Optional
import uuid
import flask
import sqlalchemy

# TODO this import here is so strange
from datetime import datetime
from jjsystem.common import exception
from jjsystem.common.operation_after_post import \
    operation_after_post_registry
from jjsystem.common.subsystem.driver import Driver


class Operation(object):

    def __init__(self, manager):
        self.manager = manager
        self.driver: Optional[Driver] = (
            manager.driver if hasattr(manager, 'driver') else None)

    def pre(self, **kwargs):
        return True

    def do(self, **kwargs):
        return True

    def post(self):
        pass

    def __call__(self, **kwargs):
        session = kwargs.pop(
            'session', self.driver.transaction_manager.session)

        if not self.pre(session=session, **kwargs):
            raise exception.PreconditionFailed()

        try:
            self.driver.transaction_manager.begin()

            result = self.do(session, **kwargs)

            self.driver.transaction_manager.commit()

            self.post()

            key = (self.manager.__class__, self.__class__)
            fn_after_post = operation_after_post_registry.get(key, None)
            if fn_after_post is not None:
                fn_after_post(self)

        except sqlalchemy.exc.IntegrityError as e:
            self.driver.transaction_manager.rollback()
            msg_info = ''.join(e.args)
            raise exception.DuplicatedEntity(msg_info)
        except Exception as e:
            self.driver.transaction_manager.rollback()
            raise e
        return result


class Create(Operation):

    def pre(self, session, **kwargs):
        if 'id' not in kwargs:
            kwargs['id'] = uuid.uuid4().hex
        if 'created_at' not in kwargs:
            kwargs['created_at'] = datetime.now()
        if 'created_by' not in kwargs:
            if flask.has_request_context():
                token_id = flask.request.headers.get('token')
                if token_id is not None:
                    self.token = self.manager.api.tokens().get(id=token_id)
                    kwargs['created_by'] = self.token.user_id

        self.entity = self.driver.instantiate(**kwargs)

        return self.entity.is_stable()

    def do(self, session, **kwargs):
        self.driver.create(self.entity, session=session)
        return self.entity


class Get(Operation):

    def pre(self, session, id, **kwargs):
        self.id = id
        return True

    def do(self, session, **kwargs):
        entity = self.driver.get(self.id, session=session)
        return entity


class List(Operation):

    def _att_kwargs_with_query(self, session, **kwargs):
        if 'order_by' in kwargs.keys():
            order_by = kwargs.get('order_by', '')
            if '.' in order_by:
                kwargs['query'] = self.manager.init_query(
                    session, order_by, self.driver.resource)
        return kwargs

    def do(self, session, **kwargs):
        kwargs = self._att_kwargs_with_query(session=session, **kwargs)
        entities = self.driver.list(session=session, **kwargs)
        total_rows = None
        if kwargs.get('require_pagination', False) is True:
            total_rows = self.driver.count(session=session, **kwargs)
            return (entities, total_rows)
        return entities


class Update(Operation):

    def pre(self, session, id, **kwargs):
        if id is None:
            raise exception.BadRequest

        self.entity = self.driver.get(id, session=session)

        self.entity.updated_at = datetime.now()
        if 'updated_by' not in kwargs:
            if flask.has_request_context():
                token_id = flask.request.headers.get('token')
                if token_id is not None:
                    self.token = self.manager.api.tokens().get(id=token_id)
                    self.entity.updated_by = self.token.user_id

        return self.entity.is_stable()

    def do(self, session, **kwargs):
        self.driver.update(self.entity, kwargs, session=session)
        return self.entity


class Delete(Operation):

    def pre(self, session, id, **kwargs):
        self.entity = self.driver.get(id, session=session)
        return True

    def do(self, session, **kwargs):
        self.driver.delete(self.entity, session=session)


class Count(Operation):

    def do(self, session, **kwargs):
        rows = self.driver.count(session=session, **kwargs)
        return rows


class ListMultipleSelection(Operation):

    def do(self, session, **kwargs):
        entities = self.driver.list_multiple_selection(
            session=session, **kwargs)
        return entities


class ActivateOrDeactivateMultipleEntities(Operation):

    def pre(self, session, **kwargs):
        domain_id = kwargs.get('domain_id', None)
        if domain_id is None:
            raise exception.BadRequest('domain_id is required.')

        active = kwargs.get('active', None)
        if active is None:
            raise exception.PreconditionFailed('active is required.')

        multiple_selection = kwargs.get('multiple_selection', None)
        if multiple_selection is None:
            raise exception.PreconditionFailed(
                'multiple_selection is required.')

        return True

    def do(self, session, **kwargs):
        result = self.driver.activate_or_deactivate_multiple_entities(
            session=session, **kwargs)
        return result




class JJSystemException(Exception):

    status = 500
    message = ''

    def __init__(self, message=None):
        if message is not None:
            self.message = message


class NotFound(JJSystemException):

    status = 404
    message = 'Entity not found'


class DuplicatedEntity(JJSystemException):

    status = 404
    message = 'Entity already exists'


class BadRequest(JJSystemException):

    status = 400
    message = 'Provided body does not represent a valid entity'


class OperationBadRequest(JJSystemException):

    status = 400
    message = 'Provided body does not provide ' + \
        'valid info for performing operation'


class BadRequestContentType(BadRequest):

    message = 'Content-Type header must be application/json'


class PreconditionFailed(BadRequest):

    message = 'One or more preconditions failed'


class FatalError(JJSystemException):

    message = 'FATAL ERROR'



from jjsystem.common.subsystem import router


class Router(router.Router):

    def __init__(self, collection, routes=[]):
        super().__init__(collection, routes)

    @property
    def routes(self):
        return [
            {
                'action': 'register',
                'method': 'POST',
                'url': '/register',
                'callback': 'register',
                'bypass': True
            },
            {
                'action': 'activate',
                'method': 'POST',
                'url': '/activate',
                'callback': 'activate',
                'bypass': True
            }
        ]



import enum

from jjsystem.database import db
from jjsystem.subsystem.file.resource import File


class QualityImage(enum.Enum):
    min = 'MIN'
    med = 'MED'
    max = 'MAX'


class Image(File, db.Model):
    MAX_ALLOWED_SIZE = (6000, 6000)

    attributes = []
    attributes += File.attributes

    id = db.Column(db.ForeignKey('file_infosys.id'), primary_key=True)
    type_image = db.Column(db.String(100), nullable=False)

    __mapper_args__ = {'polymorphic_identity': 'image'}

    def __init__(self, id, domain_id, name, type_image='', active=True,
                 created_at=None, created_by=None, updated_at=None,
                 updated_by=None, tag=None):
        super().__init__(id, domain_id, name, active, created_at, created_by,
                         updated_at, updated_by, tag)
        self.type_image = type_image

    @classmethod
    def collection(cls):
        return 'images'

    def filename_with_quality(self, quality: QualityImage):
        if quality is None:
            return '{}.{}'.format(self.id, 'jpg')
        else:
            return '{}.{}.{}'.format(self.id, quality.value, 'jpg')



from jjsystem.common import subsystem
from jjsystem.subsystem.timeline_event \
    import resource, controller, manager


subsystem = subsystem.Subsystem(resource=resource.TimelineEvent,
                                controller=controller.Controller,
                                manager=manager.Manager)



from jjsystem.common.subsystem import router


class Router(router.Router):

    def __init__(self, collection, routes=[]):
        super().__init__(collection, routes)

    @property
    def routes(self):
        return super().routes + [
            {
                'action': 'Pega as funcionalidades disponiveis',
                'method': 'GET',
                'url': self.collection_url +
                '/get_available_capability_functionalities',
                'callback': 'get_available_functionalities'
            },
            {
                'action': 'Pega as funcionalidades selecionadas',
                'method': 'GET',
                'url': self.collection_url +
                '/get_selected_capability_functionalities',
                'callback': 'get_selected_functionalities'
            },
            {
                'action': 'Cadastra as policy_functionality',
                'method': 'POST',
                'url': self.collection_url + '/create_entities',
                'callback': 'create_policy_functionalities'
            }
        ]



import json
import uuid
import unittest
import testtools

from jjsystem.tests.functional import test_base


class UserTestCase(test_base.CRUDTest,
                   testtools.TestCase):

    def load_fixtures(self):
        domain = self.post(
            resource_ref={'name': uuid.uuid4().hex, 'active': True},
            resource_name='domain',
            collection_name='domains')

        self.domain_id = domain['id']

    @property
    def resource_name(self):
        return 'user'

    @property
    def required_attributes(self):
        return ['name', 'domain_id', 'email', 'password']

    @property
    def optional_attributes(self):
        return ['active']

    @property
    def unique_attributes(self):
        return [('name', 'domain_id')]

    @property
    def hidden_attributes(self):
        return ['password']

    def new_resource_ref(self):
        return {'name': uuid.uuid4().hex,
                'domain_id': self.domain_id,
                'email': uuid.uuid4().hex,
                'password': uuid.uuid4().hex,
                'active': True}


if __name__ == '__main__':
    unittest.main()



from jjsystem.common import exception


operation_after_post_registry = dict()


def do_after_post(manager, operation):
    '''
    Decorator for register action after post based on manager operation

    Parameters:
        manager(cls): The Manager class reference
        operation(cls): The Operation class reference

    Returns:
        do_after_post(manager, operation)(fn): The original function
    '''

    def wrapper(fn):
        key = (manager, operation)
        if operation_after_post_registry.get(key):
            raise exception.JJSystemException(
                f'The operation {manager}.{operation} was already registered')

        operation_after_post_registry[key] = fn
        return fn

    return wrapper



import os
import hashlib

from sqlalchemy import or_, and_

from jjsystem.common import exception
from jjsystem.common.subsystem import manager
from jjsystem.common.subsystem import operation
from jjsystem.subsystem.functionality.resource import Functionality, FunctionalityRoute
from jjsystem.subsystem.module.resource import Module, ModuleFunctionality
from jjsystem.subsystem.user.email import TypeEmail, send_email
from jjsystem.subsystem.user.resource import User
from jjsystem.subsystem.capability.resource import Capability
from jjsystem.subsystem.capability_functionality.resource \
    import CapabilityFunctionality
from jjsystem.subsystem.capability_module.resource import CapabilityModule
from jjsystem.subsystem.domain.resource import Domain
from jjsystem.subsystem.route.resource import Route
from jjsystem.subsystem.grant.resource import Grant
from jjsystem.subsystem.policy.resource import Policy
from jjsystem.subsystem.policy_functionality.resource \
    import PolicyFunctionality
from jjsystem.subsystem.policy_module.resource import PolicyModule
from jjsystem.subsystem.role.resource import Role


class Create(operation.Create):

    def pre(self, session, **kwargs):
        self.role = self.manager.api.roles().get_role_by_name(
            role_name=Role.USER)
        kwargs.pop('settings', None)
        return super().pre(session, **kwargs)

    def do(self, session, **kwargs):
        super().do(session)
        self.manager.api.grants().create(role_id=self.role.id,
                                         user_id=self.entity.id)
        return self.entity


class Update(operation.Update):

    def do(self, session, **kwargs):
        kwargs.pop('settings', None)
        return super().do(session, **kwargs)


class UpdatePassword(operation.Update):

    def _check_password(self, password, password_db):
        if not password:
            return password_db is None
        password_hash = self.manager.hash_password(password)
        return password_hash == password_db

    def pre(self, session, id, **kwargs):
        old_password = kwargs.pop('old_password', None)
        self.password = kwargs.pop('password', None)

        if not (id and self.password and old_password):
            raise exception.BadRequest()
        super().pre(session=session, id=id)

        if not self._check_password(old_password, self.entity.password):
            raise exception.BadRequest()
        return True

    def do(self, session, **kwargs):
        self.entity.password = self.manager.hash_password(self.password)
        self.entity = super().do(session)
        return self.entity


class Restore(operation.Operation):

    def pre(self, **kwargs):
        email = kwargs.get('email', None)
        domain_name = kwargs.get('domain_name', None)
        jjsystem_reset_url = os.environ.get(
            'JJSYSTEM_RESET_URL', 'http://objetorelacional.com.br/#/reset/')
        self.reset_url = kwargs.get('reset_url', jjsystem_reset_url)

        if not (domain_name and email and self.reset_url):
            raise exception.OperationBadRequest()

        domains = self.manager.api.domains().list(name=domain_name)
        if not domains:
            raise exception.OperationBadRequest()

        self.domain = domains[0]

        users = self.manager.api.users().list(
            email=email, domain_id=self.domain.id)
        if not users:
            raise exception.OperationBadRequest()

        self.user = users[0]

        return True

    def do(self, session, **kwargs):
        self.manager.notify(
            id=self.user.id, type_email=TypeEmail.FORGOT_PASSWORD)
        # token = self.manager.api.tokens().create(user=self.user)
        # send_email(token.id, self.user, self.domain)


class Reset(operation.Update):

    def pre(self, session, id, **kwargs):
        self.password = kwargs.get('password', None)
        if not (id and self.password):
            raise exception.BadRequest()
        super().pre(session=session, id=id)
        return True

    def do(self, session, **kwargs):
        self.entity.password = self.manager.hash_password(self.password)
        self.entity = super().do(session)
        return self.entity


class Routes(operation.Operation):

    def _get_application_id(self, session, user_id):
        result_application = session.query(Domain.application_id). \
            join(User). \
            filter(User.id == user_id). \
            first()

        if not result_application.application_id:
            raise exception.BadRequest(
                'This user is not associated with any applications')

        return result_application.application_id

    def do(self, session, user_id, **kwargs):
        response = []
        application_id = self._get_application_id(session, user_id)

        query = session.query(Route). \
            join(Capability). \
            outerjoin(Policy). \
            outerjoin(Grant, Policy.role_id == Grant.role_id). \
            outerjoin(User). \
            outerjoin(Domain,
                      and_(Domain.application_id == Capability.application_id,
                           Domain.id == User.domain_id)). \
            filter(Capability.application_id == application_id,
                   Capability.active, Route.active,
                   or_(Route.bypass,
                       and_(User.id == user_id,
                            User.active, Domain.active, Grant.active,
                            Policy.active))). \
            distinct()
        projeto = kwargs.get('projeto', None)
        if projeto is not None:
            query = query.filter(
                or_(Route.projeto.is_(None), Route.projeto.is_(projeto)))

        routes = query.all()
        response += routes

        query = session.query(Route). \
            join(FunctionalityRoute). \
            join(CapabilityFunctionality, CapabilityFunctionality.functionality_id == FunctionalityRoute.functionality_id). \
            outerjoin(PolicyFunctionality). \
            outerjoin(Grant, PolicyFunctionality.role_id == Grant.role_id). \
            outerjoin(User). \
            outerjoin(Domain,
                      and_(Domain.application_id ==
                           CapabilityFunctionality.application_id,
                           Domain.id == User.domain_id)). \
            filter(CapabilityFunctionality.application_id == application_id,
                   CapabilityFunctionality.active, Route.active,
                   or_(Route.bypass,
                       and_(User.id == user_id,
                            User.active, Domain.active, Grant.active,
                            PolicyFunctionality.active))). \
            distinct()
        if projeto is not None:
            query = query.filter(
                or_(Route.projeto.is_(None), Route.projeto.is_(projeto)))

        routes = query.all()
        response += routes

        query = session.query(Route). \
            join(FunctionalityRoute). \
            join(ModuleFunctionality,
                 ModuleFunctionality.functionality_id
                 == FunctionalityRoute.functionality_id). \
            join(Module). \
            join(CapabilityModule). \
            outerjoin(PolicyModule). \
            outerjoin(Grant, PolicyModule.role_id == Grant.role_id). \
            outerjoin(User). \
            outerjoin(Domain,
                      and_(Domain.application_id ==
                           CapabilityModule.application_id,
                           Domain.id == User.domain_id)). \
            filter(CapabilityModule.application_id == application_id,
                   CapabilityModule.active, Route.active,
                   or_(Route.bypass,
                       and_(User.id == user_id,
                            User.active, Domain.active, Grant.active,
                            PolicyModule.active))). \
            distinct()
        if projeto is not None:
            query = query.filter(
                or_(Route.projeto.is_(None), Route.projeto.is_(projeto)))

        routes = query.all()
        response += routes
        response = set(response)

        return response


class Roles(operation.Operation):

    def do(self, session, user_id, **kwargs):
        roles = session.query(Role). \
            join(Grant, Grant.role_id == Role.id). \
            filter(Grant.user_id == user_id). \
            distinct(). \
            all()

        return roles


class Authorization(operation.Operation):

    def do(self, session, user_id, route, **kwargs):
        has_capabilities = session.query(User.id). \
            join(Domain). \
            join(Grant). \
            join(Role). \
            join(Policy). \
            join(Capability,
                 and_(Capability.id == Policy.capability_id,
                      Capability.application_id == Domain.application_id,
                      Capability.route_id == route.id)). \
            filter(and_(User.id == user_id,
                        or_(not route.sysadmin, Role.name == Role.SYSADMIN),
                        User.active, Domain.active, Grant.active,
                        Policy.active, Capability.active)). \
            count()

        has_capability_functionalities = session.query(User.id). \
            join(Domain). \
            join(Grant). \
            join(Role). \
            join(PolicyFunctionality). \
            join(CapabilityFunctionality,
                 and_(CapabilityFunctionality.id
                      == PolicyFunctionality.capability_functionality_id,
                      CapabilityFunctionality.application_id
                      == Domain.application_id)). \
            join(FunctionalityRoute,
                 and_(FunctionalityRoute.functionality_id
                      == CapabilityFunctionality.functionality_id,
                      FunctionalityRoute.route_id == route.id)). \
            filter(and_(User.id == user_id,
                        or_(not route.sysadmin, Role.name == Role.SYSADMIN),
                        User.active, Domain.active, Grant.active,
                        PolicyFunctionality.active,
                        CapabilityFunctionality.active)). \
            count()

        has_capability_modules = session.query(User.id). \
            join(Domain). \
            join(Grant). \
            join(Role). \
            join(PolicyModule). \
            join(CapabilityModule,
                 and_(PolicyModule.capability_module_id
                      == CapabilityModule.id,
                      CapabilityModule.application_id
                      == Domain.application_id)). \
            join(ModuleFunctionality,
                 ModuleFunctionality.module_id
                 == CapabilityModule.module_id). \
            join(FunctionalityRoute,
                 and_(FunctionalityRoute.functionality_id
                      == ModuleFunctionality.functionality_id,
                      FunctionalityRoute.route_id == route.id)). \
            filter(and_(User.id == user_id,
                        or_(not route.sysadmin, Role.name == Role.SYSADMIN),
                        User.active, Domain.active, Grant.active,
                        ModuleFunctionality.active,
                        CapabilityFunctionality.active)). \
            count()

        total = (has_capabilities + has_capability_functionalities +
                 has_capability_modules)

        return total > 0


class UploadPhoto(operation.Update):

    def pre(self, session, id, **kwargs):
        kwargs.pop('password', None)
        photo_id = kwargs.pop('photo_id', None)
        self.entity = self.manager.get(id=id)
        self.entity.photo_id = photo_id
        return self.entity.is_stable()

    def do(self, session, **kwargs):
        return super().do(session=session)


class DeletePhoto(operation.Update):

    def pre(self, session, id, **kwargs):
        kwargs.pop('password', None)
        self.entity = self.manager.get(id=id)
        self.photo_id = self.entity.photo_id
        self.entity.photo_id = None
        return self.entity.is_stable()

    def do(self, session, **kwargs):
        return super().do(session=session)

    def post(self):
        if self.photo_id:
            self.manager.api.images().delete(id=self.photo_id)


class Notify(operation.Operation):

    def _get_sysadmin(self):
        users = self.manager.list(name=User.SYSADMIN_USERNAME)
        user = users[0] if users else None
        return user

    def pre(self, session, id, type_email, **kwargs):
        self.user = self.manager.get(id=id)
        self.type_email = type_email
        if not self.user or not self.type_email:
            raise exception.BadRequest()
        return True

    def do(self, session, **kwargs):
        if self.type_email is TypeEmail.ACTIVATE_ACCOUNT:
            user_token = self._get_sysadmin()
        else:
            user_token = self.user

        self.token = self.manager.api.tokens().create(
            session=session, user=user_token)

        self.domain = self.manager.api.domains().get(id=self.user.domain_id)
        if not self.domain:
            raise exception.OperationBadRequest()

        send_email(self.type_email, self.token.id, self.user, self.domain)


class UpdateSettings(operation.Update):

    def pre(self, session, id: str, **kwargs) -> bool:
        self.settings = kwargs
        if self.settings is None or not self.settings:
            raise exception.BadRequest("Erro! There is not a setting")
        return super().pre(session=session, id=id)

    def do(self, session, **kwargs):
        result = {}
        for key, value in self.settings.items():
            new_value = self.entity.update_setting(key, value)
            result[key] = new_value
        super().do(session)

        return result


class RemoveSettings(operation.Update):

    def pre(self, session, id: str, **kwargs) -> bool:
        self.keys = kwargs.get('keys', [])
        if not self.keys:
            raise exception.BadRequest('Erro! keys are empty')
        super().pre(session, id=id)

        return self.entity.is_stable()

    def do(self, session, **kwargs):
        result = {}
        for key in self.keys:
            value = self.entity.remove_setting(key)
            result[key] = value
        super().do(session=session)

        return result


class GetUserSettingsByKeys(operation.Get):

    def pre(self, session, id, **kwargs):
        self.keys = kwargs.get('keys', [])
        if not self.keys:
            raise exception.BadRequest('Erro! keys are empty')
        return super().pre(session, id=id)

    def do(self, session, **kwargs):
        entity = super().do(session=session)
        settings = {}
        for key in self.keys:
            value = entity.settings.get(key, None)
            if value is not None:
                settings[key] = value
        return settings


class Manager(manager.Manager):

    def __init__(self, driver):
        super(Manager, self).__init__(driver)
        self.create = Create(self)
        self.update = Update(self)
        self.update_password = UpdatePassword(self)
        self.restore = Restore(self)
        self.reset = Reset(self)
        self.routes = Routes(self)
        self.upload_photo = UploadPhoto(self)
        self.delete_photo = DeletePhoto(self)
        self.notify = Notify(self)
        self.authorize = Authorization(self)
        self.roles = Roles(self)
        self.update_settings = UpdateSettings(self)
        self.remove_settings = RemoveSettings(self)
        self.get_user_settings_by_keys = GetUserSettingsByKeys(self)

    def hash_password(self, password):
        return hashlib.sha256(password.encode('utf-8')).hexdigest()



import json
import uuid
import unittest
import testtools

from jjsystem.tests.functional import test_base


class DomainTestCase(test_base.CRUDTest,
                     testtools.TestCase):

    def load_fixtures(self):
        pass

    @property
    def resource_name(self):
        return 'domain'

    @property
    def required_attributes(self):
        return ['name']

    @property
    def optional_attributes(self):
        return ['active', 'parent_id']

    @property
    def unique_attributes(self):
        return [('name',)]

    def new_resource_ref(self):
        return {'name': uuid.uuid4().hex, 'active': True, 'parent_id': None}


if __name__ == '__main__':
    unittest.main()



import json
import testtools
import uuid

import jjsystem


class JJSystemTest(object):

    def setUp(self):
        super(JJSystemTest, self).setUp()
        app = jjsystem.SystemFlask()
        app.config['TESTING'] = True
        self.app = app.test_client()

        response = self.app.post(
            '/tokens',
            data=json.dumps({'domain_name': 'default', 'username': 'admin', 'password': '123456'}),
            headers={'Content-Type': 'application/json'})

        self.token = json.loads(response.data.decode())['token']['id']

        self.load_fixtures()

    def load_fixtures(self):
        pass

    @property
    def resource_name(self):
        return NotImplemented

    @property
    def collection_name(self):
        return self.resource_name + 's'

    @property
    def required_attributes(self):
        return NotImplemented

    @property
    def optional_attributes(self):
        return []

    @property
    def unique_attributes(self):
        return NotImplemented

    @property
    def hidden_attributes(self):
        return []

    def new_resource_ref(self):
        return NotImplemented

    def check_resource(self, resource, ref=None):
        resource_copy = resource.copy()

        resource_id = resource_copy.pop('id')
        self.assertEqual(uuid.UUID(resource_id, version=4).hex,
                         resource_id)

        if ref:
            # Drop hidden attributes before comparing
            for attr in self.hidden_attributes:
                ref.pop(attr)

            if 'id' in ref:
                self.assertEqual(resource_id, ref.pop('id'))

            self.assertDictEqual(resource_copy, ref)
        else:
            attrs = (attr for attr in self.required_attributes if attr not in self.hidden_attributes)
            for attr in attrs:
                self.assertIsNotNone(resource_copy[attr])

    def post(self, resource_ref, status='201 CREATED',
             resource_name=None, collection_name=None,
             headers=None):
        if resource_name is None:
            resource_name = self.resource_name
        if collection_name is None:
            collection_name = self.collection_name
        if headers is None:
            headers = {'token': self.token,
                       'Content-Type': 'application/json'}

        if resource_ref is not None:
            resource_ref = json.dumps(resource_ref)

        response = self.app.post(
            '/' + collection_name,
            headers=headers,
            data=resource_ref)

        self.assertEqual(status, response.status)

        if status == '201 CREATED':
            return json.loads(response.data.decode())[resource_name]

    def get(self, resource_id, status='200 OK', headers=None):
        resource_path = '/' + resource_id

        if headers is None:
            headers = {'token': self.token}

        response = self.app.get(
            '/' + self.collection_name + resource_path,
            headers=headers)

        self.assertEqual(status, response.status)

        if status == '200 OK':
            return json.loads(response.data.decode())[self.resource_name]

    def list(self, status='200 OK', headers=None, **kwargs):
        for k,v in kwargs.items():
            if isinstance(v, bool):
                kwargs[k] = 'true' if v else 'false'
            elif v is None:
                kwargs[k] = 'null'
        params = '&'.join([k + '=' + str(v) for k, v in kwargs.items()])
        params = '?' + params if params else ''

        if headers is None:
            headers = {'token': self.token}

        response = self.app.get(
            '/' + self.collection_name + params,
            headers=headers)

        self.assertEqual(status, response.status)

        if status == '200 OK':
            return json.loads(response.data.decode())[self.collection_name]

    def update(self, resource_id, new_resource, status='200 OK', headers=None):
        if headers is None:
            headers = {'token': self.token,
                       'Content-Type': 'application/json'}

        response = self.app.put(
            '/%s/%s' % (self.collection_name, resource_id),
            headers=headers,
            data=json.dumps(new_resource))

        self.assertEqual(status, response.status)

        if status == '200 OK':
            return json.loads(response.data.decode())[self.resource_name]

    def delete(self, resource_id, status='204 NO CONTENT', headers=None):
        if headers is None:
            headers = {'token': self.token}
        response = self.app.delete(
            '/%s/%s' % (self.collection_name, resource_id),
            headers=headers)

        self.assertEqual(status, response.status)


class CreateTest(JJSystemTest):

    def test_create(self):
        ref = self.new_resource_ref()
        resource = self.post(ref)
        self.check_resource(resource, ref)

    def test_create_minimal_body(self):
        ref = self.new_resource_ref()
        for attr in self.optional_attributes:
            ref.pop(attr)

        resource = self.post(ref)

        # The return is valid and the provided attrs match
        self.check_resource(resource)
        # Drop hidden attributes before comparing
        for attr in self.hidden_attributes:
            ref.pop(attr)
        self.assertDictContainsSubset(ref, resource)

    def test_create_invalid_body(self):
        ref = {uuid.uuid4().hex: uuid.uuid4().hex}
        self.post(ref, '400 BAD REQUEST')

    def test_create_incomplete_body(self):
        for attr in self.required_attributes:
            ref = self.new_resource_ref()
            ref.pop(attr)
            self.post(ref, '400 BAD REQUEST')

    def test_create_without_body(self):
        self.post(None, status='400 BAD REQUEST')

    def test_create_empty_body(self):
        self.post({}, status='400 BAD REQUEST')

    def test_create_without_token(self):
        headers = {'Content-Type': 'application/json'}

        ref = self.new_resource_ref()
        resource = self.post(ref, status='401 UNAUTHORIZED',
                             headers=headers)

    def test_create_invalid_token(self):
        headers = {'token': uuid.uuid4().hex,
                   'Content-Type': 'application/json'}

        ref = self.new_resource_ref()
        resource = self.post(ref, status='401 UNAUTHORIZED',
                             headers=headers)

    def test_create_without_content_type(self):
        headers = {'token': self.token}

        ref = self.new_resource_ref()
        resource = self.post(ref, status='400 BAD REQUEST',
                             headers=headers)

    def test_create_wrong_content_type(self):
        headers = {'Content-Type': 'application/xml',
                   'token': self.token}

        ref = self.new_resource_ref()
        resource = self.post(ref, status='400 BAD REQUEST',
                             headers=headers)


class RetrieveTest(JJSystemTest):

    def test_get(self):
        ref = self.new_resource_ref()
        resource = self.post(ref)

        response = self.get(resource_id=resource['id'])
        self.check_resource(response, ref)

    def test_list(self):
        resources = []
        for ref in range(3):
            resources.append(self.post(self.new_resource_ref()))

        response = self.list()

        for resource in response:
            self.check_resource(resource)
        for resource in resources:
            self.assertIn(resource, response)

    def test_list_filtering_by_unique_attrs(self):
        resource = self.post(self.new_resource_ref())

        for attribute_set in self.unique_attributes:
            response = self.list(
                **{attr: resource[attr] for attr in attribute_set})
            self.assertItemsEqual([resource], response)

    def test_list_filtering_by_full_resource(self):
        resource = self.post(self.new_resource_ref())

        response = self.list(**resource)

        self.assertItemsEqual([resource], response)

    def test_list_without_token(self):
        self.list(headers={}, status='401 UNAUTHORIZED')

    def test_list_invalid_token(self):
        self.list(headers={'token': uuid.uuid4().hex},
                  status='401 UNAUTHORIZED')

    def test_get_without_token(self):
        ref = self.new_resource_ref()
        resource = self.post(ref)

        response = self.get(resource_id=resource['id'],
                            headers={},
                            status='401 UNAUTHORIZED')

    def test_get_invalid_token(self):
        ref = self.new_resource_ref()
        resource = self.post(ref)

        response = self.get(resource_id=resource['id'],
                            headers={'token': uuid.uuid4().hex},
                            status='401 UNAUTHORIZED')


class UpdateTest(JJSystemTest):

    def test_update(self):
        ref = self.new_resource_ref()
        resource = self.post(ref)

        updated_body = resource.copy()
        updated_body['name'] = uuid.uuid4().hex

        updated_resource = self.update(resource['id'], updated_body)

        ref.update(updated_body)
        self.check_resource(updated_resource, ref)

    def test_create_without_content_type(self):
        headers = {'token': self.token}

        resource_id = uuid.uuid4().hex
        diff_body = {}
        self.update(resource_id, diff_body, status='400 BAD REQUEST',
                             headers=headers)

    def test_create_wrong_content_type(self):
        headers = {'Content-Type': 'application/xml',
                   'token': self.token}

        resource_id = uuid.uuid4().hex
        diff_body = {}
        self.update(resource_id, diff_body, status='400 BAD REQUEST',
                             headers=headers)

    def test_update_without_token(self):
        resource_id = uuid.uuid4().hex
        diff_body = {}
        self.update(resource_id, diff_body,
                    headers={}, status='401 UNAUTHORIZED')

    def test_update_invalid_token(self):
        resource_id = uuid.uuid4().hex
        diff_body = {}
        self.update(resource_id, diff_body,
                    headers={'token': uuid.uuid4().hex},
                    status='401 UNAUTHORIZED')


class DeleteTest(JJSystemTest):

    def test_delete(self):
        ref = self.new_resource_ref()
        resource = self.post(ref)

        self.get(resource['id'])

        self.delete(resource['id'])

        self.get(resource['id'], status='404 NOT FOUND')

    def test_update_without_token(self):
        resource_id = uuid.uuid4().hex
        self.delete(resource_id,
                    headers={}, status='401 UNAUTHORIZED')

    def test_update_invalid_token(self):
        resource_id = uuid.uuid4().hex
        self.delete(resource_id,
                    headers={'token': uuid.uuid4().hex},
                    status='401 UNAUTHORIZED')


# TODO(samueldmq): Group content-type and token tests
# TODO(samueldmq): Assert what error messages actually say
# TODO(samueldmq): Use fixtures to simplify load_fixtures in classes
# TODO(samueldmq): Use local policy file for tests (with a fixture?)
# TODO(samueldmq): Test different permissions (tests for authorization)
# TODO(samueldmq): Test for token API

class CRUDTest(CreateTest, RetrieveTest, UpdateTest, DeleteTest):
    pass




from jjsystem.common.subsystem import router


class Router(router.Router):

    def __init__(self, collection, routes=[]):
        super().__init__(collection, routes)

    @property
    def routes(self):
        settings_endpoint = '/settings'
        return super().routes + [
            {
                'action': 'getApplicationRoles',
                'method': 'GET',
                'url': self.resource_url + '/roles',
                'callback': 'get_roles'
            },
            {
                'action': 'Update settings on Application',
                'method': 'PUT',
                'url': self.resource_url + settings_endpoint,
                'callback': 'update_settings',
                'bypass': False
            },
            {
                'action': 'Remove settings from Application',
                'method': 'DELETE',
                'url': self.resource_url + settings_endpoint,
                'callback': 'remove_settings',
                'bypass': False
            },
            {
                'action': 'Get settings by keys from Application',
                'method': 'GET',
                'url': self.resource_url + settings_endpoint,
                'callback': 'get_application_settings_by_keys',
                'bypass': False
            }
        ]



from datetime import datetime
from jjsystem.database import db
from jjsystem.common.subsystem import entity


class Notification(entity.Entity, db.Model):

    attributes = ['user_id', 'date', 'subject', 'body', 'read_date']
    attributes += entity.Entity.attributes

    user_id = db.Column(
        db.CHAR(32), db.ForeignKey("user.id"), nullable=False)
    date = db.Column(db.Date, nullable=False)
    subject = db.Column(db.String(50), nullable=False)
    body = db.Column(db.String(250), nullable=False)
    read_date = db.Column(db.Date, nullable=True)

    def __init__(self, id, user_id, date, subject, body, active=True,
                 read_date=None, created_at=None, created_by=None,
                 updated_at=None, updated_by=None, tag=None):
        super().__init__(id, active, created_at, created_by,
                         updated_at, updated_by, tag)
        self.user_id = user_id
        self.date = datetime.strptime(date, entity.DATETIME_FMT)
        self.subject = subject
        self.body = body
        if (read_date is not None):
            self.read_date = datetime.strptime(read_date, entity.DATETIME_FMT)



from typing import List

from jjsystem.common.input import RouteResource, InputResourceUtils
from jjsystem.common.subsystem.apihandler import Api


class BootstrapPolicies(object):

    def __init__(self, api: Api) -> None:
        self.capability_manager = api.capabilities()
        self.role_manager = api.roles()

    def execute(self, application_id: str,
                role_id: str,
                user_resources: List[RouteResource],
                sysadmin_resources: List[RouteResource],
                sysadmin_exclusive_resources: List[RouteResource]):
        capabilities_resources = InputResourceUtils. \
            join_resources(user_resources,
                           sysadmin_resources,
                           sysadmin_exclusive_resources)
        policies_resources = InputResourceUtils. \
            join_resources(sysadmin_resources,
                           sysadmin_exclusive_resources)

        self._save_capabilities(application_id, capabilities_resources)
        self._save_policies(application_id, role_id, policies_resources)

    def _save_capabilities(self, application_id: str,
                           resources: List[RouteResource]):
        data = {'resources': resources}
        self.capability_manager.create_capabilities(id=application_id,
                                                    **data)

    def _save_policies(self, application_id: str,
                       role_id: str,
                       resources: List[RouteResource]) -> None:
        data = {'resources': resources, 'application_id': application_id}
        self.role_manager.create_policies(id=role_id, **data)



from jjsystem.database import db
from sqlalchemy import UniqueConstraint
from jjsystem.common.subsystem import entity


class Route(entity.Entity, db.Model):

    attributes = ['name', 'url', 'method', 'sysadmin', 'bypass', 'projeto']
    attributes += entity.Entity.attributes

    name = db.Column(db.String(100), nullable=False)
    url = db.Column(db.String(80), nullable=False)
    method = db.Column(db.String(10), nullable=False)
    bypass = db.Column(db.Boolean(), nullable=False)
    sysadmin = db.Column(db.Boolean(), nullable=False)
    projeto = db.Column(db.Boolean(), nullable=True)

    __table_args__ = (UniqueConstraint('url', 'method', name='route_uk'),)

    def __init__(self, id, name, url, method, bypass=False, sysadmin=False,
                 projeto=None,
                 active=True, created_at=None, created_by=None,
                 updated_at=None, updated_by=None, tag=None):
        super().__init__(id, active, created_at, created_by,
                         updated_at, updated_by, tag)
        self.name = name
        self.url = url
        self.method = method
        self.bypass = bypass
        self.sysadmin = sysadmin
        self.projeto = projeto



from typing import Dict, List

from jjsystem.common.subsystem import Subsystem
from jjsystem.common.subsystem.apihandler import Api
from jjsystem.common.input import RouteResource
from jjsystem.subsystem.role.resource import Role
from jjsystem.bootstrap.roles import BootstrapRoles
from jjsystem.bootstrap.default.application import BootstrapApplication
from jjsystem.bootstrap.default.domain import BootstrapDomain
from jjsystem.bootstrap.default.user import BootstrapUser
from jjsystem.bootstrap.default.policies import BootstrapPolicies


class BootstrapDefault(object):

    def __init__(self, api: Api) -> None:
        self.bootstrap_roles = BootstrapRoles(api)
        self.bootstrap_application = BootstrapApplication(api)
        self.bootstrap_domain = BootstrapDomain(api)
        self.bootstrap_user = BootstrapUser(api)
        self.bootstrap_policies = BootstrapPolicies(api)

    def execute(self, user_resources: List[RouteResource],
                sysadmin_resources: List[RouteResource],
                sysadmin_exclusive_resources: List[RouteResource]):
        roles = self.bootstrap_roles.execute()
        role_sysadmin = self._get_role_sysadmin(roles)

        application = self.bootstrap_application.execute()
        domain = self.bootstrap_domain.execute(application.id)
        self.bootstrap_user.execute(domain.id, role_sysadmin.id)
        self.bootstrap_policies.execute(application.id,
                                        role_sysadmin.id,
                                        user_resources,
                                        sysadmin_resources,
                                        sysadmin_exclusive_resources)

    def _get_role_sysadmin(self, roles: List[Role]) -> Role:
        role = next((role for role in roles if role.name == Role.SYSADMIN),
                    None)
        if not role:
            raise Exception()
        return role



import os
import shutil
from pathlib import Path

from jjsystem.common.subsystem import operation
from jjsystem.subsystem.file import manager
from jjsystem.subsystem.image import tasks
from jjsystem.subsystem.image.resource import QualityImage
from jjsystem.subsystem.image.handler import ImageHandler
from jjsystem.common import exception


class Create(manager.Create):

    def do(self, session, **kwargs):
        super().do(session)

        # TODO (araujobd)  check a better way to improve this
        # only way to validate resolution of image was after
        # save temporary file so if exceeds should trigger rollback
        folder = self.manager.get_upload_folder(self.entity,
                                                self.entity.domain_id)
        error_message = ImageHandler.verify_size_resolution_image(
            folder, self.entity.filename)
        if error_message is not None:
            shutil.rmtree(folder)
            raise exception.BadRequest(error_message)

        return self.entity

    def post(self):
        tasks.process_image(self.upload_folder, self.entity.filename)


class Get(operation.Get):

    def pre(self, session, id, **kwargs):
        self.quality = kwargs.pop('quality', QualityImage.min)

        return super().pre(id=id, session=session)

    def do(self, session, **kwargs):
        file = super().do(session=session, **kwargs)

        folder = self.manager.get_upload_folder(file, file.domain_id)
        filename = file.filename_with_quality(self.quality)

        existingFile = Path(f'{folder}/{filename}')
        if existingFile.is_file() is False and self.quality is not None:
            filename = file.filename_with_quality(None)
            existingFile = Path(f'{folder}/{filename}')

        if existingFile.is_file() is False:
            raise exception.JJSystemException('File not found!')
        else:
            return folder, filename


class Delete(operation.Delete):

    def post(self):
        # TODO(fdoliveira) Put this in worker
        folder = self.manager.get_upload_folder(self.entity,
                                                self.entity.domain_id)
        shutil.rmtree(folder)


class Manager(manager.Manager):
    ALLOWED_EXTENSIONS = ['jpg', 'jpeg', 'png', 'gif']

    def __init__(self, driver):
        super().__init__(driver)
        self.create = Create(self)
        self.get = Get(self)
        self.delete = Delete(self)

    def get_upload_folder(self, entity, domain_id):
        base_folder = self._get_base_folder()
        entity_name = type(entity).__name__
        folder = os.path.join(base_folder,
                              entity_name,
                              entity.type_image,
                              self.OPTIONAL_FOLDER,
                              domain_id,
                              entity.id)
        if not os.path.exists(folder):
            os.makedirs(folder)
        return folder



from jjsystem.common import subsystem
from jjsystem.common.subsystem import controller
from jjsystem.subsystem.project_cost \
    import resource, manager

subsystem = subsystem.Subsystem(resource=resource.ProjectCost,
                                manager=manager.Manager,
                                controller=controller.Controller)



from flask.globals import current_app

from jjsystem.celery import celery, decide_on_run
from jjsystem.common import exception
from jjsystem.subsystem.user.email import TypeEmail


@decide_on_run
@celery.task(autoretry_for=(exception.NotFound,),
             default_retry_delay=5,
             retry_kwargs={'max_retries': 3})
def send_email(user_id: str) -> None:
    api = current_app.api_handler.api()
    users_manager = api.users()
    return users_manager.notify(id=user_id,
                                type_email=TypeEmail.ACTIVATE_ACCOUNT)



from datetime import datetime
from jjsystem.database import db
from jjsystem.common.subsystem import entity


class Token(entity.Entity, db.Model):

    attributes = ['user_id', 'natureza']
    attributes += entity.Entity.attributes

    user_id = db.Column(db.CHAR(32), db.ForeignKey("user.id"), nullable=False)
    natureza = db.Column(db.String(20), nullable=True)

    def __init__(self, id, user_id, natureza=None,
                 active=True, created_at=datetime.now(),
                 created_by=user_id, updated_at=None,
                 updated_by=None, tag=None):
        super().__init__(id, active, created_at, created_by,
                         updated_at, updated_by, tag)
        self.user_id = user_id
        self.natureza = natureza



from jjsystem.database import db
from jjsystem.common.subsystem import entity


class File(entity.Entity, db.Model):

    # TODO(fdoliveira) Check in another databases
    # This is required because Oracle dont accept FILE for table name
    __tablename__ = 'file_infosys'

    attributes = ['domain_id', 'name']
    attributes += entity.Entity.attributes

    domain_id = db.Column(
        db.CHAR(32), db.ForeignKey("domain.id"), nullable=True)
    name = db.Column(db.String(255), nullable=True)

    def __init__(self, id, domain_id, name,
                 active=True, created_at=None, created_by=None,
                 updated_at=None, updated_by=None, tag=None):
        super().__init__(id, active, created_at, created_by,
                         updated_at, updated_by, tag)
        self.domain_id = domain_id
        self.name = name

    @property
    def extension(self):
        name = self.name.split('.')
        return name[-1]

    @property
    def filename(self):
        return '{}.{}'.format(self.id, self.extension)



from jjsystem.common.subsystem import operation, manager
from jjsystem.subsystem.constant_for_calculation.resource \
    import ConstantForCalculation
from jjsystem.subsystem.application.resource import Application
from jjsystem.common.subsystem.pagination import Pagination


class List(operation.List):

    def do(self, session, **kwargs):
        query = session.query(ConstantForCalculation). \
            join(Application, Application.id == # noqa
                 ConstantForCalculation.application_id)
        query = query.distinct()
        dict_compare = {"application.": Application}
        kwargs['query'] = query
        kwargs['dict_compare'] = dict_compare
        return super().do(session=session, **kwargs)


class Manager(manager.Manager):

    def __init__(self, driver):
        super(Manager, self).__init__(driver)
        self.list = List(self)



from sqlalchemy import orm, UniqueConstraint

from jjsystem.database import db
from jjsystem.common.subsystem import entity


class CapabilityFunctionality(entity.Entity, db.Model):

    attributes = ['functionality_id', 'application_id']
    attributes += entity.Entity.attributes

    functionality_id = db.Column(
        db.CHAR(32), db.ForeignKey("functionality.id"), nullable=False)
    functionality = orm.relationship(
        'Functionality',
        backref=orm.backref('capability_functionality_functionality'))
    application_id = db.Column(
        db.CHAR(32), db.ForeignKey("application.id"), nullable=False)
    application = orm.relationship('Application', backref=orm.backref(
        'capability_functionality_application'))

    __table_args__ = (
        UniqueConstraint(
            'functionality_id', 'application_id',
            name='cp_functionality_id_application_id_uk'),)

    def __init__(self, id, functionality_id, application_id,
                 active=True, created_at=None, created_by=None,
                 updated_at=None, updated_by=None, tag=None):
        super().__init__(id, active, created_at, created_by,
                         updated_at, updated_by, tag)
        self.functionality_id = functionality_id
        self.application_id = application_id

    @classmethod
    def individual(cls):
        return 'capability_functionality'

    @classmethod
    def collection(cls):
        return 'capability_functionalities'



import flask

from jjsystem.common import exception, utils
from jjsystem.common.subsystem import controller


class Controller(controller.Controller):

    def __init__(self, manager, resource_wrap, collection_wrap):
        super(Controller, self).__init__(
            manager, resource_wrap, collection_wrap)

    def get_token_id(self):
        return flask.request.headers.get('token')

    def get_token(self, token_id):
        return self.manager.api.tokens().get(id=token_id)

    def add_functionalities(self, id):
        try:
            data = flask.request.get_json()

            module = self.manager.add_functionalities(id=id, **data)

            response = {'module': module.to_dict({'functionalities': {}})}
        except exception.JJSystemException as exc:
            return flask.Response(response=exc.message,
                                  status=exc.status)

        return flask.Response(response=utils.to_json(response),
                              status=201,
                              mimetype="application/json")

    def rm_functionalities(self, id):
        try:
            data = flask.request.get_json()

            module = self.manager.rm_functionalities(id=id, **data)

            response = {'module': module.to_dict({'functionalities': {}})}
        except exception.JJSystemException as exc:
            return flask.Response(response=exc.message,
                                  status=exc.status)

        return flask.Response(response=utils.to_json(response),
                              status=201,
                              mimetype="application/json")

    def get_available_functionalities(self, id):
        filters = self._filters_parse()

        try:
            filters = self._parse_list_options(filters)
            functionalities, total_rows = self.manager\
                .get_available_functionalities(id=id, **filters)

            collection = self._entities_to_dict(
                functionalities, self._get_include_dicts(filters))

            response = {
                'module': {
                    'id': id,
                    'functionalities': collection
                }
            }

            page = filters.get('page', None)
            page_size = filters.get('page_size', None)

            if total_rows is not None:
                response.update(
                    {'pagination': {'page': int(page),
                                    'page_size': int(page_size),
                                    'total': total_rows}})
        except exception.JJSystemException as exc:
            return flask.Response(response=exc.message,
                                  status=exc.status)

        return flask.Response(response=utils.to_json(response),
                              status=200,
                              mimetype="application/json")

    def get_selected_functionalities(self, id):
        filters = self._filters_parse()

        try:
            filters = self._parse_list_options(filters)
            functionalities, total_rows = self.manager\
                .get_selected_functionalities(id=id, **filters)

            collection = self._entities_to_dict(
                functionalities, self._get_include_dicts(filters))

            response = {
                'module': {
                    'id': id,
                    'functionalities': collection
                }
            }

            page = filters.get('page', None)
            page_size = filters.get('page_size', None)

            if total_rows is not None:
                response.update(
                    {'pagination': {'page': int(page),
                                    'page_size': int(page_size),
                                    'total': total_rows}})
        except exception.JJSystemException as exc:
            return flask.Response(response=exc.message,
                                  status=exc.status)

        return flask.Response(response=utils.to_json(response),
                              status=200,
                              mimetype="application/json")



from sqlalchemy import func, or_, and_
from jjsystem.subsystem.application.resource import Application
from typing import List
from jjsystem.common import exception
from jjsystem.common.input import RouteResource, InputResource, \
    InputResourceUtils
from jjsystem.common.subsystem import operation, manager
from jjsystem.subsystem.capability.resource import Capability
from jjsystem.subsystem.policy.resource import Policy
from jjsystem.subsystem.role.resource import Role
from jjsystem.subsystem.route.resource import Route
from jjsystem.common.subsystem.pagination import Pagination
from sqlalchemy.sql import text


class Create(operation.Create):

    def pre(self, session, **kwargs) -> bool:
        self.exceptions = kwargs.pop('exceptions', [])

        return super().pre(session, **kwargs)

    def do(self, session, **kwargs):
        super().do(session)
        self.manager.create_user_capabilities_and_policies(id=self.entity.id,
                                                           session=session)
        if self.entity.name != Application.DEFAULT:
            self.manager.create_admin_capabilities_and_policies(
                id=self.entity.id, session=session, exceptions=self.exceptions)
        return self.entity


class ListManager(operation.List):

    def do(self, session, **kwargs):
        not_default = kwargs.pop('not_default', False)
        query = session.query(Application)
        if not_default is True:
            query = query.filter(Application.name != Application.DEFAULT)
        kwargs['query'] = query
        return super().do(session=session, **kwargs)


class CreateUserCapabilitiesAndPolicies(operation.Operation):

    def pre(self, session, id, **kwargs) -> bool:
        self.application_id = id
        self.user_resources = self.manager.bootstrap_resources.USER
        self.role_id = self.manager.api.roles().\
            get_role_by_name(role_name=Role.USER).id

        return True

    def do(self, session, **kwargs):
        self.resources = {'resources': self.user_resources}
        self.manager.api.capabilities().create_capabilities(
            id=self.application_id, **self.resources)

        self.resources['application_id'] = self.application_id
        self.manager.api.roles().create_policies(id=self.role_id,
                                                 **self.resources)


class CreateAdminCapabilitiesAndPolicies(operation.Operation):

    def _map_routes(self, routes: List[Route]) -> List[RouteResource]:
        resources = [(route.url, route.method) for route in routes]
        return resources

    def _filter_resources(self, all_resources: List[RouteResource],
                          exceptions_resources: List[RouteResource],
                          sysadmin_exclusive_resources: List[RouteResource],
                          user_resources: List[RouteResource]) \
            -> List[RouteResource]:
        resources = InputResourceUtils.diff_resources(
            all_resources, sysadmin_exclusive_resources, user_resources,
            exceptions_resources)
        return resources

    def pre(self, session, id, exceptions: List[InputResource] = [], **kwargs):
        self.application_id = id
        exceptions_resources = InputResourceUtils.parse_resources(exceptions)

        routes = self.manager.api.routes().list(active=True)
        routes_resources = self._map_routes(routes)

        self.admin_role_id = self.manager.api.roles().\
            get_role_by_name(role_name=Role.ADMIN).id

        self.admin_resources = self._filter_resources(
            routes_resources, exceptions_resources,
            self.manager.bootstrap_resources.SYSADMIN_EXCLUSIVE,
            self.manager.bootstrap_resources.USER)
        return True

    def do(self, session, **kwargs):
        self.resources = {'resources': self.admin_resources}
        self.manager.api.capabilities().create_capabilities(
            id=self.application_id, **self.resources)

        self.resources['application_id'] = self.application_id
        self.manager.api.roles().create_policies(id=self.admin_role_id,
                                                 **self.resources)


class CreateCapabilitiesWithExceptions(operation.Operation):

    def _filter_resources(self, routes: List[Route],
                          exceptions: List[RouteResource],
                          sysadmin_exclusive_resources: List[RouteResource],
                          user_resources: List[RouteResource]) \
            -> List[RouteResource]:
        all_resources = [(route.url, route.method) for route in routes]
        resources = InputResourceUtils.diff_resources(
            all_resources, sysadmin_exclusive_resources, user_resources,
            exceptions)
        return resources

    def pre(self, session, id: str, **kwargs):
        self.application_id = id
        exceptions = kwargs.get('exceptions', None)

        if not self.application_id or exceptions is None:
            raise exception.BadRequest()

        routes = self.manager.api.routes().list(active=True)
        exceptions_resources = InputResourceUtils.parse_resources(exceptions)
        self.resources = self.\
            _filter_resources(routes,
                              exceptions_resources,
                              self.manager.bootstrap_resoures.
                              SYSADMIN_EXCLUSIVE,
                              self.manager.bootstrap_resources.USER)

        return self.driver.get(id, session=session) is not None

    def do(self, session, **kwargs):
        data = {'resources', self.resources}
        self.manager.api.capabilities().\
            create_capabilities(id=self.application_id, **data)


class GetRoles(operation.Operation):

    def pre(self, session, id, **kwargs):
        self.application_id = id
        return self.driver.get(id, session=session) is not None

    def do(self, session, **kwargs):
        roles = session.query(Role). \
            join(Policy). \
            join(Capability). \
            filter(and_(Capability.application_id == self.application_id,
                        Role.name != Role.USER)). \
            distinct()
        return roles


class UpdateSettings(operation.Update):

    def pre(self, session, id: str, **kwargs) -> bool:
        self.settings = kwargs
        if self.settings is None or not self.settings:
            raise exception.BadRequest("Erro! There is not a setting")
        return super().pre(session=session, id=id)

    def do(self, session, **kwargs):
        result = {}
        for key, value in self.settings.items():
            new_value = self.entity.update_setting(key, value)
            result[key] = new_value
        super().do(session)

        return result


class RemoveSettings(operation.Update):

    def pre(self, session, id: str, **kwargs) -> bool:
        self.keys = kwargs.get('keys', [])
        if not self.keys:
            raise exception.BadRequest('Erro! keys are empty')
        super().pre(session, id=id)

        return self.entity.is_stable()

    def do(self, session, **kwargs):
        result = {}
        for key in self.keys:
            value = self.entity.remove_setting(key)
            result[key] = value
        super().do(session=session)

        return result


class GetApplicationSettingsByKeys(operation.Get):

    def pre(self, session, id, **kwargs):
        self.keys = kwargs.get('keys', [])
        if not self.keys:
            raise exception.BadRequest('Erro! keys are empty')
        return super().pre(session, id=id)

    def do(self, session, **kwargs):
        entity = super().do(session=session)
        settings = {}
        for key in self.keys:
            value = entity.settings.get(key, None)
            if value is not None:
                settings[key] = value
        return settings


class Manager(manager.Manager):

    def __init__(self, driver):
        super().__init__(driver)
        self.create = Create(self)
        self.list = ListManager(self)
        self.create_user_capabilities_and_policies = \
            CreateUserCapabilitiesAndPolicies(self)
        self.create_admin_capabilities_and_policies = \
            CreateAdminCapabilitiesAndPolicies(self)
        self.get_roles = GetRoles(self)
        self.update_settings = UpdateSettings(self)
        self.remove_settings = RemoveSettings(self)
        self.get_application_settings_by_keys = \
            GetApplicationSettingsByKeys(self)



from jjsystem.common.subsystem.apihandler import Api
from typing import Dict, List

from jjsystem.common.subsystem import Subsystem
from jjsystem.common.input import RouteResource
from jjsystem.bootstrap.default import BootstrapDefault
from jjsystem.bootstrap.routes import BootstrapRoutes


class Bootstrap(object):

    def __init__(self, api: Api,
                 subsystems: Dict[str, Subsystem],
                 user_resources: List[RouteResource],
                 sysadmin_resources: List[RouteResource],
                 sysadmin_exclusive_resources: List[RouteResource]):
        self.user_resources = user_resources
        self.sysadmin_resources = sysadmin_resources
        self.sysadmin_exclusive_resources = sysadmin_exclusive_resources

        self.routes = BootstrapRoutes(subsystems, api)
        self.default = BootstrapDefault(api)
        self.application_manager = api.applications()

    def execute(self):
        self.routes.execute()

        if not self.application_manager.list():
            self.default.execute(self.user_resources,
                                 self.sysadmin_resources,
                                 self.sysadmin_exclusive_resources)



import uuid
from jjsystem.common.subsystem.pagination import Pagination
from jjsystem.common.subsystem.entity import DATE_FMT
from datetime import datetime as datetime1
import datetime as datetime2
from typing import Any, Type

from jjsystem.common import exception
from sqlalchemy import func, or_, and_
from sqlalchemy.orm import exc
from sqlalchemy.sql import text

from jjsystem.common.subsystem.transaction_manager import TransactionManager


class Driver(object):

    def __init__(self, resource: Type[Any],
                 transaction_manager: TransactionManager) -> None:
        self.resource = resource
        self.transaction_manager = transaction_manager

    def removeId(self, entity_aux):
        new_id = uuid.uuid4().hex

        if entity_aux.get('id') is not None:
            new_id = entity_aux.pop('id')

        return new_id

    def instantiate(self, **kwargs):
        try:
            embedded = {}
            for attr in self.resource.embedded():
                if attr not in kwargs:
                    raise Exception(
                        f'O campo embedded {attr} é obrigatório em ' +
                        f'{self.resource.__name__}.')
                embedded.update({attr: kwargs.pop(attr)})

            instance = self.resource(**kwargs)

            for attr in embedded:
                value = embedded[attr]
                var = getattr(self.resource, attr)
                # TODO(samueldmq): is this good enough? should we discover it?
                mapped_attr = {self.resource.individual() + '_id': instance.id}
                if isinstance(value, list):
                    setattr(instance, attr, [var.property.mapper.class_(
                        id=self.removeId(ref), **dict(ref, **mapped_attr))
                        for ref in value])
                else:
                    # TODO(samueldmq): id is inserted here. it is in the
                    # manager for the entities. do it all in the resource
                    # contructor
                    setattr(instance, attr, var.property.mapper.class_(
                        id=uuid.uuid4().hex, **dict(value, **mapped_attr)))
        except Exception as exec:
            # TODO(samueldmq): replace with specific exception
            message = ''.join(exec.args)
            raise exception.BadRequest(message)

        return instance

    def create(self, entity, session):
        if not entity.is_stable():
            raise exception.PreconditionFailed()
        session.add(entity)
        session.flush()

    def update(self, entity, data, session):
        # try:
        #     entity = self.get(id, session)
        # except exc.NoResultFound:
        #     raise exception.NotFound()

        for attr in self.resource.embedded():
            if attr in data:
                value = data.pop(attr)
                var = getattr(self.resource, attr)
                # TODO(samueldmq): is this good enough? should we discover it?
                mapped_attr = {self.resource.individual() + '_id': id}
                if isinstance(value, list):
                    setattr(entity, attr, [var.property.mapper.class_(
                        id=self.removeId(ref), **dict(ref, **mapped_attr))
                        for ref in value])
                else:
                    # TODO(samueldmq): id is inserted here. it is in the
                    # manager for the entities. do it all in the resource
                    # contructor
                    setattr(entity, attr, var.property.mapper.class_(
                        id=uuid.uuid4().hex, **dict(value, **mapped_attr)))

        for key, value in data.items():
            if hasattr(entity, key):
                try:
                    setattr(entity, key, value)
                except AttributeError:
                    raise exception.BadRequest(
                        f'Error! The attribute {key} is read only')
            else:
                raise exception.BadRequest(
                    f'Error! The attribute {key} not exists')

        if not entity.is_stable():
            raise exception.PreconditionFailed()
        session.flush()
        return entity

    def delete(self, entity, session):
        session.delete(entity)
        session.flush()

    def get(self, id, session):
        try:
            query = session.query(self.resource).filter_by(id=id)
            result = query.one()
        except exc.NoResultFound:
            raise exception.NotFound()

        return result

    def list(self, session, **kwargs):
        dict_compare = kwargs.pop('dict_compare', None)
        # pega a query ou monta uma
        query = kwargs.pop('query', session.query(self.resource))

        # aplica filtros na classe principal
        query = self.apply_filters(query, self.resource, **kwargs)
        # aplica filtros nas classes dos joins
        query = self.apply_filters_includes(query, dict_compare, **kwargs)
        # aplica filtros de intervalo de tempo em colunas date ou datetime
        query = self.apply_filter_multiple_de_ate_with_timezone(
            query=query, resource=self.resource, **kwargs)
        # aplica order_by se "order_by" tiver sido passado
        query = self.apply_order_by(query, **kwargs)
        # aplica paginação se "page" e "page_size" tiver sido passado
        query = self.apply_pagination(query, **kwargs)

        result = query.all()

        if not (type(result) is list):
            result = list(map(lambda x: x[0], result))

        return result

    def count(self, session, **kwargs):
        try:
            # TODO(JogeSilva): improve filtering so as not to ignore parameters
            # that are attributes of an entity to include
            # pega a query ou monta uma
            query = kwargs.pop('query', session.query(self.resource.id))
            rows = self.apply_filters(query, self.resource, **kwargs).count()
            result = rows
        except exc.NoResultFound:
            raise exception.NotFound()

        return result

    def activate_or_deactivate_multiple_entities(self, session, **kwargs):
        active = kwargs.pop('active', None)

        entities = self.list_multiple_selection(session, **kwargs)
        key = 'active'

        for entity in entities:
            if hasattr(entity, key):
                setattr(entity, key, active)
            else:
                raise exception.BadRequest(
                    f'Error! The attribute {key} not exists.')

            if not entity.is_stable():
                raise exception.PreconditionFailed()
        session.flush()
        return entities

    def list_multiple_selection(self, session, **kwargs):
        '''
            fields that can be passed in kwargs:

            dict_compare
            resource
            query

            the possibility of passing these fields as parameters makes
            the function generic.
        '''

        dict_compare = kwargs.pop('dict_compare', {})
        query_in_kwargs = False
        multiple_selection = kwargs.get('multiple_selection', None)
        if multiple_selection is None:
            raise exception.BadRequest('multiple_selection is required.')

        resource_filtro = self.resource
        resource = kwargs.pop('resource', None)
        if resource is not None:
            resource_filtro = resource

        if 'query' in kwargs.keys():
            query = kwargs.pop('query', None)
            query_in_kwargs = True
        else:
            query = session.query(resource_filtro)

        selected_list = multiple_selection.get('selected_list', [])
        unselected_list = multiple_selection.get('unselected_list', [])

        if len(selected_list) > 0:
            query = query.filter(resource_filtro.id.in_(selected_list))
        else:
            if len(unselected_list) > 0:
                query = query.filter(resource_filtro.id.not_in(unselected_list))
            query = self.apply_filters(query, resource_filtro, **kwargs)
            query = self.apply_filters_includes(query, dict_compare, **kwargs)
            query = self.apply_filter_multiple_de_ate_with_timezone(
                query=query, resource=self.resource, **kwargs)

        query = query.distinct()
        result = query.all()

        if query_in_kwargs is True:
            result = list(map(lambda x: x[0], result))

        return result

    def apply_filters(self, query, resource, **kwargs):
        for k, v in kwargs.items():
            if hasattr(resource, k):
                if k == 'tag':
                    values = v
                    if len(v) > 0 and v[0] == '#':
                        values = v[1:]
                    values = values.split(',')
                    filter_tags = []
                    for value in values:
                        filter_tags.append(
                            getattr(resource, k)
                            .like('%#' + str(value) + ' %'))
                    query = query.filter(or_(*filter_tags))
                elif isinstance(v, str) and '%' in v:
                    normalize = func.jjsystem_normalize
                    query = query.filter(normalize(getattr(resource, k))
                                         .ilike(normalize(v)))
                else:
                    query = query.filter(getattr(resource, k) == v)

        return query

    def apply_order_by(self, query, **kwargs):
        tablename = f'"{self.resource.__tablename__}"'
        order_by_default = f'{tablename}.id'
        order_by = kwargs.get('order_by', None)
        # se não tiver sido passado um order_by,
        # irá usar o order_by default
        if order_by is None:
            query = query.order_by(text(order_by_default))
        else:
            # se foi passado um order_by ai iremos normalizar ele
            order_by_splited = order_by.split(',')
            order_by_final = ''
            for item in order_by_splited:
                if '.' not in item:
                    order_by_final += f'{tablename}.{item},'
                else:
                    order_by_final += f'{item},'

            query = query.order_by(text(order_by_final[:-1]))
        return query

    def apply_pagination(self, query, **kwargs):
        page_size = int(kwargs.get('page_size', 9999999))
        page = int(kwargs.get('page', 0))

        query = query.limit(page_size)
        query = query.offset(page * page_size)

        return query

    def apply_filters_includes(self, query, dict_compare, **kwargs):
        if type(dict_compare) is not dict:
            return query
        for id, resource in dict_compare.items():
            for k, v in kwargs.items():
                if id in k and hasattr(resource, k.split('.')[-1]):
                    k = k.split('.')[-1]
                    isinstance_aux = isinstance(v, str)

                    if k == 'tag':
                        # TODO(JorgeSilva): definir o caractere para split
                        values = v
                        if len(v) > 0 and v[0] == '#':
                            values = v[1:]
                        values = values.split(',')
                        filter_tags = []
                        for value in values:
                            filter_tags.append(
                                getattr(resource, k)
                                .like('%#' + str(value) + ' %'))
                        query = query.filter(or_(*filter_tags))
                    elif isinstance_aux and self.__isdate(v):
                        day, next_day = self.__get_day_and_next_day(v)
                        query = query.filter(
                            and_(
                                or_(getattr(resource, k) < next_day,
                                    getattr(resource, k) == None),  # noqa
                                or_(getattr(resource, k) >= day,
                                    getattr(resource, k) == None)))  # noqa
                    elif isinstance_aux and '%' in v:
                        normalize = func.jjsystem_normalize
                        query = query.filter(normalize
                                             (getattr(resource, k))
                                             .ilike(normalize(v)))
                    else:
                        query = query.filter(getattr(resource, k) == v)

        return query

    def with_pagination(self, **kwargs):
        require_pagination = kwargs.get('require_pagination', False)
        page = kwargs.get('page', None)
        page_size = kwargs.get('page_size', None)

        if (page and page_size is not None) and require_pagination is True:
            return True
        return False

    def __isdate(self, data, format="%Y-%m-%d"):
        res = True
        try:
            res = bool(datetime1.strptime(data, format))
        except ValueError:
            res = False
        return res

    def __get_day_and_next_day(self, data, format="%Y-%m-%d"):
        day = datetime1.strptime(data, format)
        next_day = day + datetime2.timedelta(days=1)
        return (day, next_day)

    def _convert_de_ate(self, **kwargs):
        de = kwargs.get('de', None)
        ate = kwargs.get('ate', None)
        inicio = None
        fim = None

        if de and ate:
            try:
                inicio = datetime1.strptime(de.replace(' ', '+'), '%Y-%m-%d%z')
                fim = datetime1.strptime(ate.replace(' ', '+'), '%Y-%m-%d%z') \
                    + datetime2.timedelta(days=1)

            except Exception:
                inicio = datetime1.strptime(de, DATE_FMT)
                fim = datetime1.strptime(ate, DATE_FMT) +\
                    datetime2.timedelta(days=1)

        return (inicio, fim)

    # função criada para filtrar na listagem de uma entidade mais
    # de um campo do tipo Date ou DateTime
    def apply_filter_multiple_de_ate_with_timezone(
            self, query, resource, **kwargs):
        attributes_filter = kwargs.get('attributes_filter', '')
        attributes_filter = attributes_filter.split(',')
        de_filter = kwargs.get('de_filter', '')
        de_filter = de_filter.split(',')
        ate_filter = kwargs.get('ate_filter', '')
        ate_filter = ate_filter.split(',')

        tamanho = len(attributes_filter)

        if not (tamanho == len(de_filter) and tamanho == len(ate_filter)):
            return query

        for i in range(tamanho):
            attribute = attributes_filter[i]
            data = {
                'de': de_filter[i],
                'ate': ate_filter[i]
            }
            (de, ate) = self._convert_de_ate(**data)
            if de and ate:
                if hasattr(resource, attribute):
                    query = query.filter(
                        and_(getattr(resource, attribute) >= de,
                             getattr(resource, attribute) < ate))
        return query

    def adjust_dinamic_order_by(self, resource: Type[Any]):
        if self.order_by is not None:
            order_by_ajusted = ''
            table_name = resource.__tablename__
            order_by_post_split = self.order_by.split(',')
            for item in order_by_post_split:
                if '.' in item:
                    order_by_ajusted += f'{item},'
                else:
                    order_by_ajusted += f'{table_name}.{item},'

            self.order_by = order_by_ajusted[:-1]



import flask

from jjsystem.common import exception, utils
from jjsystem.common.subsystem import controller


class Controller(controller.Controller):

    def __init__(self, manager, resource_wrap, collection_wrap):
        super(Controller, self).__init__(
            manager, resource_wrap, collection_wrap)

    def get_nextval(self, id):
        data = flask.request.get_json()

        try:
            if 'name' not in data or 'table_id' not in data:
                raise exception.BadRequest(
                    'ERROR! "name" or "table_id" not defined')

            response = self.manager.get_nextval(
                id=id,
                table_id=data['table_id'],
                name=data['name'])
        except exception.JJSystemException as exc:
            return flask.Response(response=exc.message,
                                  status=exc.status)

        response = {'nextval': response}

        return flask.Response(response=utils.to_json(response),
                              status=200,
                              mimetype="application/json")



import os
import importlib


def import_subsystems(path, package):
    dirs = [
        d for d in os.listdir(path) if os.path.isdir(os.path.join(path, d))]
    modules = [importlib.import_module(package + '.' + m) for m in dirs]
    return [m.subsystem for m in modules if hasattr(m, 'subsystem')]


# TODO(samueldmq): import here or at destination directly ?
all = import_subsystems(
    os.path.dirname(os.path.realpath(__file__)), 'jjsystem.subsystem')



from typing import Any, Type, Optional
from jjsystem.common import exception
from jjsystem.common.exception import BadRequest


class Pagination(object):

    def __init__(self, page: Optional[int], page_size: Optional[int]) -> None:
        self.page = page
        self.page_size = page_size

    @classmethod
    def get_pagination(cls, resource: Type[Any], **kwargs):
        try:
            page = kwargs.pop('page', None)
            page = int(page) if page is not None else None
            page_size = kwargs.pop('page_size', None)
            page_size = int(page_size) if page_size is not None else None

            name_pagination_column = 'pagination_column'

            # if order_by is None and page is not None and page_size is not None \
            #         and hasattr(resource, name_pagination_column):
            #     order_by = getattr(resource, name_pagination_column)

            if page_size is not None and page_size <= 0:
                raise exception.BadRequest(
                    'page_size must be greater than to zero')
            if page is not None and page < 0:
                raise exception.BadRequest(
                    'page must be greater than or equal to zero')

            # cls._validate_order_by(order_by, resource)
        except BadRequest as br:
            raise exception.BadRequest(br.message)
        except ValueError:
            raise exception.BadRequest('page or page_size is invalid')

        return cls(page=page, page_size=page_size)

    # def _validate_order_by(order_by: Optional[str], resource: Type[Any]):
    #     if order_by is None:
    #         return None
    #     order_by_post_split = order_by.split(',')
    #     for item in order_by_post_split:
    #         splited_item = item.split()

    #         if len(splited_item) < 3 and len(splited_item) > 0:
    #             if len(splited_item) == 1 or (splited_item[1] == 'asc' or
    #                splited_item[1] == 'desc'):

    #                 attr = splited_item[0]
    #                 if hasattr(resource, attr):
    #                     table_name = resource.__tablename__
    #                     columns = resource.metadata.tables[table_name].columns\
    #                         ._all_columns
    #                     columns_name = list(
    #                         map(lambda item: item.name, columns))
    #                     if attr not in columns_name:
    #                         raise exception.BadRequest(f'{attr} is \
    #                             not a column in table {table_name}')
    #                 else:
    #                     raise exception.BadRequest(f'{attr} is \
    #                             not a attribute in class {resource.__name__}')
    #             else:
    #                 raise exception.BadRequest('Each order_by item must be the \
    #                     attribute name and optionally "asc" (sort ascending) or\
    #                     "desc" (sort descending)')
    #         else:
    #             raise exception.BadRequest('each order_by item must be one or\
    #                 two words at most')

    # def adjust_order_by(self, resource: Type[Any]):
    #     if self.order_by is not None:
    #         order_by_ajusted = ''
    #         table_name = resource.__tablename__
    #         order_by_post_split = self.order_by.split(',')
    #         for item in order_by_post_split:
    #             order_by_ajusted += f'{table_name}.{item},'

    #         self.order_by = order_by_ajusted[:-1]

    # def adjust_order_by_no_distinct(self, resource: Type[Any]):
    #     if self.order_by is not None:
    #         order_by_ajusted = ''
    #         table_name = resource.__tablename__
    #         order_by_post_split = self.order_by.split(',')
    #         for item in order_by_post_split:
    #             item_aux = item.split(' ')
    #             if 'VARCHAR' in str(getattr(resource, item_aux[0]).type):
    #                 if ' ' in item:
    #                     order_by_ajusted += (
    #                         'ordenar_sem_formatacao(' +
    #                         f'{table_name}.{item_aux[0]}) {item_aux[1]},')
    #                 else:
    #                     order_by_ajusted += (
    #                         f'ordenar_sem_formatacao({table_name}.{item}),')
    #             else:
    #                 order_by_ajusted += f'{table_name}.{item},'

    #         self.order_by = order_by_ajusted[:-1]

    # def adjust_dinamic_order_by(self, resource: Type[Any]):
    #     if self.order_by is not None:
    #         order_by_ajusted = ''
    #         table_name = resource.__tablename__
    #         order_by_post_split = self.order_by.split(',')
    #         for item in order_by_post_split:
    #             if '.' in item:
    #                 order_by_ajusted += f'{item},'
    #             else:
    #                 order_by_ajusted += f'{table_name}.{item},'

    #         self.order_by = order_by_ajusted[:-1]



from jjsystem.common.subsystem import operation, manager
from jjsystem.subsystem.project_cost.resource import ProjectCost
from jjsystem.common.subsystem.pagination import Pagination


class List(operation.List):

    def do(self, session, **kwargs):
        query = session.query(ProjectCost)
        query = self.manager.apply_filters(
            query, ProjectCost, **kwargs)

        query = self.manager.apply_filter_de_ate_with_timezone(
            ProjectCost, query, **kwargs)
        query = query.distinct()

        total_rows = None
        if self.manager.with_pagination(**kwargs):
            total_rows = query.count()

        pagination = Pagination.get_pagination(
            ProjectCost, **kwargs)
        if pagination.order_by is not None:
            pagination.adjust_order_by(ProjectCost)
        query = self.driver.apply_pagination(query, pagination)
        result = query.all()

        return (result, total_rows)


class GetCostByNameMostRecent(operation.List):

    def do(self, session, **kwargs):
        kwargs['order_by'] = 'created_at desc'
        costs = super().do(session, **kwargs)
        if len(costs) > 0:
            return costs[0].cost
        return None


class Manager(manager.Manager):

    def __init__(self, driver):
        super(Manager, self).__init__(driver)
        self.list = List(self)
        self.get_cost_by_name_most_recent = GetCostByNameMostRecent(self)



from sqlalchemy import orm, UniqueConstraint

from jjsystem.database import db
from jjsystem.common.subsystem import entity


class ConstantForCalculation(entity.Entity, db.Model):

    attributes = ['application_id', 'table_name', 'p_processos',
                  'p_acessos_db']
    attributes += entity.Entity.attributes

    application_id = db.Column(
        db.CHAR(32), db.ForeignKey("application.id"), nullable=False)
    application = orm.relationship('Application', backref=orm.backref(
        'constants_for_calculation_application'))
    table_name = db.Column(db.String(60), nullable=False)
    p_processos = db.Column(db.Numeric(5), nullable=False)
    p_acessos_db = db.Column(db.Numeric(5), nullable=False)

    __table_args__ = (UniqueConstraint(
        'application_id', 'table_name',
        name='constant_for_calculation_application_id_table_name_uk'),)

    def __init__(self, id, application_id, table_name, p_processos=1,
                 p_acessos_db=1,
                 active=True, created_at=None, created_by=None,
                 updated_at=None, updated_by=None, tag=None):
        super().__init__(id, active, created_at, created_by,
                         updated_at, updated_by, tag)
        self.application_id = application_id
        self.table_name = table_name
        self.p_processos = p_processos
        self.p_acessos_db = p_acessos_db

    @classmethod
    def individual(cls):
        return 'constant_for_calculation'



import flask
import os

from jjsystem import database
from jjsystem import request
from jjsystem import subsystem as subsystem_module
from jjsystem import scheduler
from jjsystem import celery
from jjsystem.bootstrap import Bootstrap
from jjsystem.common.subsystem.apihandler import ApiHandler
from jjsystem.common.input import InputResourceUtils
from jjsystem.system import System
from jjsystem._version import version as jjsystem_version

from jjsystem.resources import SYSADMIN_EXCLUSIVE_POLICIES, \
    SYSADMIN_RESOURCES, USER_RESOURCES


system = System('jjsystem',
                subsystem_module.all,
                USER_RESOURCES,
                SYSADMIN_RESOURCES,
                SYSADMIN_EXCLUSIVE_POLICIES)


class SystemFlask(flask.Flask):

    request_class = request.Request

    def __init__(self, *args, **kwargs):
        super().__init__(__name__, static_folder=None)

        self.configure()
        self.configure_commands()
        self.init_database()
        self.after_init_database()

        system_list = [system] + list(kwargs.values()) + list(args)

        subsystem_list, self.user_resources, self.sysadmin_resources, \
            self.sysadmin_exclusive_resources = self._parse_systems(
                system_list)

        self.subsystems = {s.name: s for s in subsystem_list}

        self.api_handler = self.inject_dependencies()
        self.__validate_routes(self.subsystems)

        for subsystem in self.subsystems.values():
            self.register_blueprint(subsystem)

        # Add version in the root URL
        self.add_url_rule('/', view_func=self.version, methods=['GET'])

        self.before_request(
            request.RequestManager(self.api_handler).before_request)

    def configure(self):
        self.config['BASEDIR'] = os.path.abspath(os.path.dirname(__file__))
        self.config['SQLALCHEMY_TRACK_MODIFICATIONS'] = True
        self.config['SQLALCHEMY_DATABASE_URI'] = 'sqlite:///:memory:'
        self.config['USE_WORKER'] = False

    def configure_commands(self):
        bootstrap_decorator = self.cli.command(name='bootstrap',
                                               help='Perform bootstrap')
        bootstrap_command = bootstrap_decorator(self.bootstrap)
        self.cli.add_command(bootstrap_command)

    def init_database(self):
        database.db.init_app(self)
        with self.app_context():
            database.db.create_all()

    def after_init_database(self):
        pass

    def version(self):
        return jjsystem_version

    def schedule_jobs(self):
        pass

    def inject_dependencies(self) -> ApiHandler:
        bootstrap_resources = {
            'USER': self.user_resources,
            'SYSADMIN': self.sysadmin_resources,
            'SYSADMIN_EXCLUSIVE': self.sysadmin_exclusive_resources
        }
        api_handler = ApiHandler(self.subsystems, bootstrap_resources)

        for subsystem in self.subsystems.values():
            subsystem.api = api_handler.api

        return api_handler

    def __validate_routes(self, subsystems):
        errors = []
        for subsystem in subsystems.values():
            errors = errors + subsystem.validate_routes()
        if errors:
            for i in errors:
                print(i)  # TODO change to logger
            raise Exception(*errors)

    def bootstrap(self):
        """Bootstrap the system.

        - routes;
        - TODO(samueldmq): sysadmin;
        - default domain with admin and capabilities.

        """

        with self.app_context():
            api = self.api_handler.api()
            Bootstrap(api,
                      self.subsystems,
                      self.user_resources,
                      self.sysadmin_resources,
                      self.sysadmin_exclusive_resources).\
                execute()

    def _parse_systems(self, systems):
        user_resources = []
        sysadmin_resources = []
        sysadmin_exclusive_resources = []
        subsystems = []
        for system in systems:
            subsystems += system.subsystems
            user_resources += system.user_resources
            sysadmin_resources += system.sysadmin_resources
            sysadmin_exclusive_resources += system.sysadmin_exclusive_resources

        utils = InputResourceUtils
        user_resources = utils.remove_duplicates(user_resources)
        sysadmin_resources = utils.remove_duplicates(sysadmin_resources)
        sysadmin_exclusive_resources = utils.remove_duplicates(
            sysadmin_exclusive_resources)

        return (subsystems, user_resources,
                sysadmin_resources, sysadmin_exclusive_resources)

    def configure_celery(self):
        use_worker = self.config.get('USE_WORKER', False)
        if use_worker:
            celery.init_celery(self)

    def init_scheduler(self):
        self.scheduler = scheduler.Scheduler()
        self.schedule_jobs()
